#!/usr/bin/env python
#
# convert bch to xml
#
# Author: Ronald Tanner
# 2016-10-20
#
import sys
import re
import io
from .bch import Reader
import xml.etree.ElementTree as el
import xml.dom.minidom


def usage():
    print("femagtools.bchxml <bchfilename>")
    sys.exit(1)


def list_to_xml(tag, l):
    '''
    Turn a list into XML
    '''

    if not isinstance(tag, str):
        try:
            tag = str(tag)
        except Exception as e:
            print("bchxml.list_to_xml():  Can not convert tag {} to str!".format(tag))
            pass

    elem = el.Element(re.sub(r'^(\d+)$', r'w\1', tag))
    for v in l:
        if v is not None:  # check against None explicitly to keep ZERO values. thomas.maier/OSWALD
            if isinstance(v, list):
                elem.append(list_to_xml('val', v))
            elif isinstance(v, dict):
                elem.append(dict_to_xml('val', v))
            else:
                child = el.Element('val')
                child.text = str(v)
                elem.append(child)
    return elem


def dict_to_xml(tag, d):
    '''
    Turn a simple dict of key/value pairs into XML
    '''

    if not isinstance(tag, str):
        try:
            tag = str(tag)
        except Exception as e:
            print("bchxml.dict_to_xml():  Can not convert tag {} to str!".format(tag))
            pass

    elem = el.Element(re.sub(r'^(\d+)$', r'w\1', tag))
    for key, val in d.items():
        if val is not None:  # check against None explicitly to keep ZERO values. thomas.maier/OSWALD
            if isinstance(val, dict):
                elem.append(dict_to_xml(key, val))
            elif isinstance(val, list):
                elem.append(list_to_xml(key, val))
            else:
                child = el.Element(key)
                child.text = str(val)
                elem.append(child)
    return elem


if __name__ == "__main__":
    if len(sys.argv) != 2:
        usage()

    filename = sys.argv[1]
    bchresults = Reader()
    with io.open(filename, encoding='latin1', errors='ignore') as f:
        bchresults.read(f.readlines())

    reparsed = xml.dom.minidom.parseString(el.tostring(
        dict_to_xml("bch", bchresults), method='xml'))

    with io.open(filename.split('.')[0]+'.xml',
                 mode='w', encoding='utf-8') as f:
        f.write(reparsed.toprettyxml(indent='  '))



# -*- coding: utf-8 -*-
"""manage a geometry with
    lines, circles and arcs built from DXF

  NOTE: This code is in highly experimental state.
        Use at your own risk.

  Author: Ronald Tanner
    Date: 2017/07/06
"""
from __future__ import print_function
import numpy as np
import logging
from .shape import Element, Circle, Line, Shape
from .functions import point, points_are_close
from .functions import alpha_angle, normalise_angle, middle_angle
from .functions import line_m, line_n, mirror_point
from .functions import within_interval, part_of_circle
logger = logging.getLogger('femagtools.geom')


#############################
#          Machine          #
#############################

class Machine(object):
    def __init__(self, geom, center, radius, startangle=0, endangle=0):
        self.geom = geom
        self.center = center
        self.radius = radius
        self.startangle = startangle
        self.endangle = endangle
        self.mirror_orig_geom = None
        self.mirror_geom = None
        self.mirror_startangle = 0.0
        self.mirror_endangle = 0.0
        self.part = self.part_of_circle()
        self.airgaps = []
        self.airgap_radius = 0.0
        self.airgap2_radius = 0.0
        self.geom.center = center
        self.previous_machine = None

    def __str__(self):
        return "Machine\n" + \
               "Center: ({})\n".format(self.center) + \
               "Radius: {}\n".format(self.radius) + \
               "Angles: Start={}, End={}\n".format(self.startangle,
                                                   self.endangle) + \
               "Mirror: {}\n".format(self.mirror_geom is not None)

    def is_a_machine(self):
        return self.radius > 0.0

    def is_in_middle(self):
        return self.radius > 0.0 and \
            points_are_close(self.center, [0.0, 0.0], 1e-8)

    def is_full(self):
        return self.radius > 0.0 and \
            self.startangle == 0.0 and self.endangle == 0.0

    def is_half_up(self):
        return self.radius > 0.0 and \
               np.isclose(self.startangle, 0.0, 1e-8) and \
               (np.isclose(self.endangle, np.pi, 1e-8) or
                np.isclose(self.endangle, -np.pi, 1e-8))

    def is_half_down(self):
        return self.radius > 0.0 and \
               (np.isclose(self.endangle, np.pi, 1e-8) or
                np.isclose(self.endangle, -np.pi, 1e-8)) and \
               np.isclose(self.startangle, 0.0, 1e-8)

    def is_half_left(self):
        return self.radius > 0.0 and \
               np.isclose(self.startangle, np.pi/2, 1e-8) and \
               np.isclose(self.endangle, -np.pi/2, 1e-8)

    def is_half_right(self):
        return self.radius > 0.0 and \
               np.isclose(self.startangle, -np.pi/2, 1e-8) and \
               np.isclose(self.endangle, np.pi/2, 1e-8)

    def is_half(self):
        return self.is_half_up() or self.is_half_down() or \
               self.is_half_left() or self.is_half_right()

    def is_quarter(self):
        return self.radius > 0.0 and \
               np.isclose(alpha_angle(self.startangle, self.endangle), np.pi/2)

    def move_to_middle(self):
        if not self.is_in_middle():
            if self.radius > 0.0:
                offset = [-(self.center[0]), -(self.center[1])]
                self.geom.move(offset)
                self.set_center(0.0, 0.0)
                self.geom.clear_cut_lines()
                return True
        return False

    def set_center(self, x, y):
        self.center[0] = x
        self.center[1] = y
        self.geom.center = [x, y]

    def set_radius(self, radius):
        self.radius = radius

    def set_kind(self, kind):
        self.geom.kind = kind

    def set_inner(self):
        self.geom.is_inner = True

    def set_outer(self):
        self.geom.is_outer = True

    def clear_cut_lines(self):
        self.geom.clear_cut_lines()
        if self.mirror_geom is not None:
            self.mirror_geom.clear_cut_lines()

    def copy(self, startangle, endangle,
             airgap=False, inside=True, split=False):
        if airgap and self.airgap_radius > 0.0:
            if inside:
                if self.airgap2_radius > 0.0:
                    new_radius = min(self.airgap_radius, self.airgap2_radius)
                else:
                    new_radius = self.airgap_radius
                clone = self.geom.copy_shape(self.center,
                                             self.radius,
                                             startangle, endangle,
                                             0.0, new_radius, split)
            else:
                new_radius = self.radius
                gap_radius = max(self.airgap_radius, self.airgap2_radius)
                clone = self.geom.copy_shape(self.center,
                                             self.radius,
                                             startangle, endangle,
                                             gap_radius, self.radius+9999,
                                             split)

            circ = Circle(Element(center=self.center,
                                  radius=self.airgap_radius))
            clone.add_cut_line(circ)
        else:
            new_radius = self.radius
            clone = self.geom.copy_shape(self.center, self.radius,
                                         startangle, endangle, 0.0,
                                         self.radius+9999, split)

        if not np.isclose(normalise_angle(startangle),
                          normalise_angle(endangle), 0.0):
            start_p = point(self.center, self.radius+5, startangle)
            start_line = Line(Element(start=self.center, end=start_p))
            clone.add_cut_line(start_line)

            end_p = point(self.center, self.radius+5, endangle)
            end_line = Line(Element(start=self.center, end=end_p))
            clone.add_cut_line(end_line)

        if not np.isclose(alpha_angle(startangle, endangle), 2*np.pi):
            return Machine(clone, self.center, new_radius,
                           startangle, endangle)
        else:
            # Der Originalwinkel bleibt bestehen
            return Machine(clone, self.center, new_radius,
                           self.startangle, self.endangle)

    def full_copy(self):
        clone = self.geom.copy_shape(self.center, self.radius,
                                     0.0, 2*np.pi,
                                     0.0, self.radius+9999)
        return clone.get_machine()

    def copy_mirror(self, startangle, midangle, endangle):
        geom1 = self.geom.copy_shape(self.center,
                                     self.radius,
                                     startangle,
                                     midangle,
                                     0.0,
                                     self.radius+9999,
                                     rtol=1e-08,
                                     atol=1e-08)
        geom2 = self.geom.copy_shape(self.center,
                                     self.radius,
                                     midangle,
                                     endangle,
                                     0.0,
                                     self.radius+9999,
                                     rtol=1e-08,
                                     atol=1e-08)

        machine = Machine(geom1, self.center, self.radius,
                          startangle, midangle)
        machine.mirror_orig_geom = self.geom
        machine.mirror_geom = geom2
        machine.mirror_geom.center = self.center
        machine.mirror_startangle = midangle
        machine.mirror_endangle = endangle
        return machine

    def undo_mirror(self):
        assert(self.is_mirrored())
        assert(self.previous_machine)
        self.previous_machine.set_minmax_radius()
        # self.previous_machine.complete_hull()
        self.previous_machine.create_auxiliary_lines()
        self.previous_machine.set_kind(self.geom.kind)
        return self.previous_machine

    def rotate_to(self, new_startangle):
        if np.isclose(new_startangle, self.startangle):
            return

        if points_are_close(self.center, [0.0, 0.0]):
            angle = new_startangle - self.startangle
            self.geom.rotate(angle)
            self.startangle = new_startangle
            self.endangle += angle

    def airgap(self, correct_airgap=0.0, correct_airgap2=0.0, atol=0.1):
        logger.debug('locking for airgap')
        self.airgap_radius = 0.0
        self.airgap2_radius = 0.0

        if np.isclose(self.radius, 0.0):
            logger.debug('no radius')
            return

        if correct_airgap < 0:
            logger.debug('no airgap')
            return False  # no airgap

        self.airgaps = []
        airgaps = self.geom.detect_airgaps(self.center,
                                           self.startangle,
                                           self.endangle, atol)

        alpha = alpha_angle(self.startangle, self.endangle)

        if len(airgaps) == 1:
            self.airgaps = airgaps
        elif len(airgaps) > 0:
            lower_radius = -1.0
            upper_radius = -1.0

            for g in airgaps:
                if np.isclose(g[0], upper_radius):
                    if not self.geom.delete_airgap_circle(self.center,
                                                          lower_radius,
                                                          upper_radius,
                                                          g[1],
                                                          alpha):
                        lower_radius = g[0]
                else:
                    if lower_radius > 0.0:
                        self.airgaps.append((lower_radius, upper_radius))
                    lower_radius = g[0]
                upper_radius = g[1]
            self.airgaps.append((lower_radius, upper_radius))

        if len(self.airgaps) > 0:
            airgap_candidates = []
            for g in self.airgaps:
                gap_radius = round((g[0]+g[1])/2.0, 6)
                gap_dist = g[1] - g[0]
                circle = Circle(Element(center=self.center,
                                        radius=gap_radius))
                ok, borders = self.geom.is_airgap(self.center,
                                                  self.radius,
                                                  self.startangle,
                                                  self.endangle,
                                                  circle, atol)
                if not ok:
                    logger.error("FATAL: No Airgap with radius {}".
                                 format(gap_radius))
                    print("FATAL: No Airgap with radius {}".
                          format(gap_radius))
                    self.geom.airgaps.append(circle)
                    return True  # bad exit

                airgap_candidates.append((borders, circle, gap_dist))
                self.geom.airgaps.append(circle)

                if correct_airgap > 0.0:
                    if within_interval(correct_airgap, g[0], g[1], 0.0, 0.0):
                        self.airgap_radius = gap_radius  # ok

                if correct_airgap2 > 0.0:
                    if within_interval(correct_airgap2, g[0], g[1], 0.0, 0.0):
                        self.airgap2_radius = gap_radius  # ok

        if correct_airgap > 0.0 and self.airgap_radius == 0.0:
            logger.error("No airgap with radius {} found"
                         .format(correct_airgap))
            self.show_airgap_candidates(airgap_candidates, False)
            return True  # bad exit

        if correct_airgap2 > 0.0 and self.airgap2_radius == 0.0:
            logger.error("No airgap2 with radius {} found"
                         .format(correct_airgap2))
            self.show_airgap_candidates(airgap_candidates, False)
            return True  # bad exit

        if len(self.airgaps) == 0:
            logger.debug('No airgap found')
            return False  # no airgaps found

        if self.airgap_radius > 0.0:
            return False  # correct airgap set

        gaps = [c for b, c, d in airgap_candidates if b == 0]
        if len(gaps) == 1:  # one candidate without border intersection
            self.airgap_radius = gaps[0].radius
            return False  # ok

        if len(airgap_candidates) == 1:  # one candidate found
            self.airgap_radius = airgap_candidates[0][1].radius
            return False  # ok

        self.airgap_radius = self.show_airgap_candidates(airgap_candidates,
                                                         True)
        return False  # ok

    def show_airgap_candidates(self, airgap_candidates, get_one):
        if get_one:
            logger.info("{} airgap candidate(s) found:"
                        .format(len(airgap_candidates)))
        else:
            print("{} airgap candidate(s) found:"
                  .format(len(airgap_candidates)))
        dist = 999
        circle = None
        for b, c, d in airgap_candidates:
            if get_one:
                logger.info(" --- {}   (width={})".format(c.radius, d))
            else:
                print(" --- {}   (width={})".format(c.radius, d))
            if d < dist:
                dist = d
                circle = c

        if get_one:
            logger.info("airgap {} prefered".format(circle.radius))
            return circle.radius

        print("Use options --airgap/--airgap2 <float> to specify")
        return None

    def has_airgap(self):
        return self.airgap_radius > 0.0

    def airgap_x(self):
        return self.airgap_radius

    def airgap_y(self):
        return 0.1

    def part_of_circle(self, pos=3):
        return part_of_circle(self.startangle, self.endangle, pos)

    def delete_center_circle(self):
        gaps = self.geom.get_gaplist(self.center)
        if len(gaps) < 2:
            return

        first_gap = gaps[0][1]
        second_gap = gaps[1][0]
        if first_gap != second_gap:
            return

        first_dist = gaps[0][1]
        second_dist = gaps[1][1] - gaps[1][0]
        if first_dist < 1.0:
            if second_dist / first_dist > 10.0:
                self.geom.delete_circle((0.0, 0.0), first_dist)

    def repair_hull(self):
        logger.info('repair_hull')
        if self.is_full() and not self.has_airgap():
            self.delete_center_circle()

        if self.startangle == self.endangle:
            return

        self.geom.repair_hull_line(self.center, self.startangle)
        self.geom.repair_hull_line(self.center, self.endangle)

        if self.mirror_geom:
            self.mirror_geom.repair_hull_line(self.center,
                                              self.mirror_startangle)
            self.mirror_geom.repair_hull_line(self.center,
                                              self.mirror_endangle)

    def set_minmax_radius(self):
        self.geom.set_minmax_radius(self.center)

    def complete_hull(self, is_inner, is_outer):
        logger.info('complete_hull')
        start_corners = self.geom.complete_hull_line(self.center,
                                                     self.startangle)
        end_corners = self.geom.complete_hull_line(self.center,
                                                   self.endangle)

        if start_corners[0].is_new_point or end_corners[0].is_new_point:
            self.geom.complete_hull_arc(self.center,
                                        self.startangle, start_corners[0],
                                        self.endangle, end_corners[0],
                                        self.geom.min_radius)

        if start_corners[1].is_new_point or end_corners[1].is_new_point:
            self.geom.complete_hull_arc(self.center,
                                        self.startangle, start_corners[1],
                                        self.endangle, end_corners[1],
                                        self.geom.max_radius)

        self.set_alfa_and_corners()

    def create_auxiliary_lines(self):
        self.geom.create_auxiliary_lines(self.endangle)

    def set_alfa_and_corners(self):
        self.geom.start_corners = self.geom.get_corner_nodes(self.center,
                                                             self.startangle)
        self.geom.end_corners = self.geom.get_corner_nodes(self.center,
                                                           self.endangle)
        self.geom.alfa = alpha_angle(self.startangle, self.endangle)

        if self.mirror_geom is not None:
            self.geom.mirror_corners = self.geom.end_corners

    def is_mirrored(self):
        return self.mirror_geom is not None

    def find_symmetry(self, sym_tolerance):
        if self.radius <= 0.0:
            return False

        return self.geom.find_symmetry(self.center, self.radius,
                                       self.startangle, self.endangle,
                                       sym_tolerance)

    def get_symmetry_slice(self):
        if not self.geom.has_symmetry_area():
            return None

        machine_slice = self.copy(self.geom.symmetry_startangle(),
                                  self.geom.symmetry_endangle())
        machine_slice.clear_cut_lines()
        machine_slice.repair_hull()
        machine_slice.rotate_to(0.0)
        machine_slice.set_alfa_and_corners()
        return machine_slice

    def get_symmetry_mirror(self):
        if self.part == 1:
            # a complete machine
            startangle = 0.0
            endangle = 0.0
            midangle = np.pi
        else:
            startangle = self.startangle
            endangle = self.endangle
            midangle = middle_angle(self.startangle, self.endangle)

        machine_mirror = self.copy_mirror(startangle, midangle, endangle)
        machine_mirror.clear_cut_lines()
        machine_mirror.repair_hull()
        machine_mirror.set_alfa_and_corners()
        if machine_mirror.check_symmetry_graph(0.001, 0.05):
            machine_mirror.previous_machine = self
            return machine_mirror
        return None

    def get_symmetry_part(self):
        if self.is_mirrored():
            return self.part/2
        else:
            return self.part

    def check_symmetry_graph(self, rtol, atol):
        axis_p = point(self.center, self.radius, self.mirror_startangle)
        axis_m = line_m(self.center, axis_p)
        axis_n = line_n(self.center, axis_m)

        def is_node_available(n, nodes):
            mirror_p = mirror_point(n, self.center, axis_m, axis_n)
            for p in nodes:
                if points_are_close(p, mirror_p, rtol, atol):
                    return True
            return False

        def get_hit_factor(nodes1, nodes2):
            hit = 0
            if not nodes1:
                return 0.0

            for n in nodes1:
                if is_node_available(n, nodes2):
                    hit += 1
            return float(hit) / len(nodes1)

        hit_factor1 = get_hit_factor(self.geom.g.nodes(),
                                     self.mirror_geom.g.nodes())
        if hit_factor1 < 0.9:
            return False  # not ok

        hit_factor2 = get_hit_factor(self.mirror_geom.g.nodes(),
                                     self.geom.g.nodes())
        if hit_factor2 < hit_factor1:
            return False  # not ok
        return True

    def sync_with_counterpart(self, cp_machine):
        self.geom.sym_counterpart = cp_machine.get_symmetry_part()
        self.geom.sym_part = self.get_symmetry_part()
        cp_machine.geom.sym_counterpart = self.get_symmetry_part()
        cp_machine.geom.sym_part = cp_machine.get_symmetry_part()

    def search_subregions(self):
        self.geom.search_subregions()

    def delete_tiny_elements(self, mindist):
        self.geom.delete_tiny_elements(mindist)



#!/usr/bin/env python
#
# read a dxf file and create a plot or fsl file
#
# Author: Ronald Tanner
# Date: 2016/01/24
#

import sys
import femagtools
from femagtools.dxfsl.converter import converter
import argparse
import logging
import logging.config
import networkx as nx
import matplotlib

logger = logging.getLogger(__name__)


def usage(name):
    print("Usage: ", name,
          " [-h] [--help]")


#############################
#            Main           #
#############################

if __name__ == "__main__":
    loglevel = logging.INFO

    argparser = argparse.ArgumentParser(
        description='Process DXF file and create a plot or FSL file.')
    argparser.add_argument('dxfile',
                           help='name of DXF file')
    argparser.add_argument('--inner',
                           help='name of inner element',
                           dest='inner',
                           default='inner')
    argparser.add_argument('--outer',
                           help='name of outer element',
                           dest='outer',
                           default='outer')
    argparser.add_argument('--rotor',
                           help='rotor without airgap in/out',
                           dest='rotor',
                           default='')
    argparser.add_argument('--stator',
                           help='stator without airgap in/out',
                           dest='stator',
                           default='')
    argparser.add_argument('-a', '--airgap',
                           help='correct airgap',
                           dest='airgap',
                           type=float,
                           default=0.0)
    argparser.add_argument('--airgap2',
                           help='correct airgap',
                           dest='airgap2',
                           type=float,
                           default=0.0)
    argparser.add_argument('-t', '--symtol',
                           help='absolut tolerance to find symmetry axis',
                           dest='sym_tolerance',
                           type=float,
                           default=0.001)
    argparser.add_argument('--mindist',
                           help='minimal distance of spline control-points',
                           dest='mindist',
                           type=float,
                           default=0.0)
    argparser.add_argument('--rtol',
                           help='relative tolerance (pickdist)',
                           dest='rtol',
                           type=float,
                           default=1e-03)
    argparser.add_argument('--atol',
                           help='absolut tolerance (pickdist)',
                           dest='atol',
                           type=float,
                           default=1e-03)
    argparser.add_argument('-s', '--split',
                           help='split intersections',
                           dest='split',
                           action="store_true")
    argparser.add_argument('-p', '--plot',
                           help='show plots',
                           dest='show_plots',
                           action="store_true")
    argparser.add_argument('-f', '--fsl',
                           help='create fsl',
                           dest='write_fsl',
                           action="store_true")
    argparser.add_argument('-v', '--view',
                           help='show a view only',
                           dest='view',
                           action="store_true")
    argparser.add_argument('-d', '--debug',
                           help='print debug information',
                           dest='debug',
                           action="store_true")
    argparser.add_argument('-l', '--log',
                           help='print debug information',
                           dest='debug',
                           action="store_true")
    argparser.add_argument('--version',
                           help='show version of some packages',
                           dest='version',
                           action="store_true")

    args = argparser.parse_args()

    if args.debug:
        loglevel = logging.DEBUG
    logging.basicConfig(level=loglevel,
                        format='%(asctime)s %(message)s')

    if args.version:
        logger.info("femagtools version: %s", femagtools.__version__)
        logger.info("networkx version: %s", nx.__version__)
        logger.info("matplotlib version: %s", matplotlib.__version__)
        sys.exit(0)

    if args.airgap > 0.0:
        if args.airgap2 > 0.0:
            logger.info("Airgap is set from {} to {}".
                        format(args.airgap, args.airgap2))
        else:
            logger.info("Airgap is set to {}".format(args.airgap))

    part = ()
    if args.stator:
        if args.rotor:
            logger.error("Stator or Rotor expected")
            sys.exit(1)
        part = ('stator', args.stator)
    elif args.rotor:
        part = ('rotor', args.rotor)
    if part:
        if args.airgap:
            logger.info('airgap in stator or rotor not possible')
            sys.exit(1)
        args.airgap = -1  # no airgap
        if part[1] not in ('in', 'out'):
            logger.info('{} has to be defined in/out'.format(part[0]))
            sys.exit(1)

    if not args.write_fsl:
        if not args.show_plots:
            args.write_fsl = True

    converter(args.dxfile,  # DXF-Filename
              rtol=args.rtol,    # relative pickdist toleranz
              atol=args.atol,    # absolute pickdist toleranz
              symtol=args.sym_tolerance,
              mindist=args.mindist,
              split=args.split,
              inner_name=args.inner,
              outer_name=args.outer,
              part=part,
              airgap=args.airgap,
              airgap2=args.airgap2,
              view_only=args.view,
              show_plots=args.show_plots,
              write_fsl=args.write_fsl,
              debug_mode=args.debug)



"""
    femagtools.docker
    ~~~~~~~~~~~~~~~~~

    Running FEMAG on Docker Swarm


"""
import os
import shutil
import json
import logging
import threading
import femagtools.femag
import femagtools.job
try:
    from queue import Queue
except ImportError:
    from Queue import Queue  # python 2.7


logger = logging.getLogger(__name__)


def get_port_binding():
    """returns list with dict(HostIp, HostPort) of all
    running femag containers:
      [{'HostPort': '25555', 'HostIp': '0.0.0.0'}, 
       {'HostPort': '15555', 'HostIp': '0.0.0.0'}, 
       {'HostPort': '5555', 'HostIp': '0.0.0.0'}]
    """
    import docker
    client = docker.from_env()
    return [c.attrs['NetworkSettings']['Ports']['5555/tcp'][0]
            for c in client.containers.list(
                    filters={'label': 'org.label-schema.name=profemag/femag'})]


def publish_receive(message):
    """handle messages from femag publisher"""
    topic, content = message  # "femag_log" + text
    # topics: femag_log, progress, file_modified,
    #   model_image, calc_image, field_image, babs_image, demag_image, color_scale
    if topic == 'femag_log' or topic == 'progress':
        logger.info("%s: %s", topic, content.strip())
    else:
        logger.info('%s: len %d', topic, len(content.strip()))


class AsyncFemag(threading.Thread):
    def __init__(self, queue, workdir, port, host):
        threading.Thread.__init__(self)
        self.queue = queue
        self.workdir = workdir
        self.container = femagtools.femag.ZmqFemag(
            workdir,
            port, host)
        
    def run(self):
        """execute femag fsl task in task directory"""
        while True:
            task = self.queue.get()
            if task is None:
                break
            fslfile = os.path.join(task.directory, task.fsl_file)
            # TODO: there must be a better way to set the workdir
            workdir = '/'.join(task.directory.split('/')[-3:])
            logger.info('Docker task %s %s workdir %s',
                        task.id, task.fsl_file, workdir)
            fslcmds = ['save_model(close)',
                       "chdir('{}')".format(workdir)]
                
            with open(fslfile) as f:
                fslcmds += f.readlines()
            r = [json.loads(s)
                 for s in self.container.send_fsl(
                         '\n'.join(fslcmds), publish_receive)[:2]]
            try:
                if r[0]['status'] == 'ok':
                    task.status = 'C'
                else:
                    task.status = 'X'
            except:
                task.status = 'X'
            logger.info("Finished %s", r)
            self.queue.task_done()
        
    
class Engine(object):

    """The Docker Engine

       execute Femag-Simulations with docker
    """
    def __init__(self, hosts=[]):
        self.hosts = hosts
        self.femag_port = int(os.environ.get('FEMAG_PORT', 5555))

    def create_job(self, workdir):
        """Create a FEMAG :py:class:`CloudJob`

        Args:
            workdir (str): The workdir where the calculation files are stored

        Return:
            job (:class:`Job`)
        """
        if not self.hosts:
            raise ValueError("empty host list")
        self.queue = Queue()
        self.async_femags = [AsyncFemag(self.queue,
                                        workdir,
                                        self.femag_port,
                                        h)
                             for h in self.hosts]
        
        self.job = femagtools.job.Job(workdir)
        return self.job

    def submit(self):
        """Starts the FEMAG calculation(s) as Docker containers

        Return:
            number of started tasks (int)
        """
        for async_femag in self.async_femags:
            async_femag.start()

        for task in self.job.tasks:
            self.queue.put(task)

        return len(self.job.tasks)

    def join(self):
        """Wait until all calculations are finished

        Return:
            list of all calculations status (C = Ok, X = error) (:obj:`list`)
        """
        # block until all tasks are done
        self.queue.join()

        # stop workers
        for _ in self.async_femags:
            self.queue.put(None)
        for async_femag in self.async_femags:
            async_femag.join()
            
        return [t.status for t in self.job.tasks]


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    engine = Engine()
    job = engine.create_job('/tmp/tar/docker-femag')
    for _ in range(3):
        t = job.add_task()
        t.add_file('femag.fsl',
                   content=[''])
    engine.submit()
    status = engine.join()
    print("Status {}".format(status))
    for t in engine.job.tasks:
        print(t.directory)
    



# -*- coding: utf-8 -*-
"""
    femagtools.femag
    ~~~~~~~~~~~~~~~~

    Running FEMAG



"""
import subprocess
import os
import glob
import logging
try:
    import zmq
except ImportError:
    pass
import sys
import json
import io
import femagtools.mcv
import femagtools.airgap as ag
import femagtools.fsl
import femagtools.ntib as ntib
import femagtools.config as cfg
import time
import platform
import re
from threading import Thread

logger = logging.getLogger(__name__)


BCHEXT = '.BATCH' if sys.platform.startswith('linux') else '.BCH'  # win32


class FemagError(Exception):
    pass


class BaseFemag(object):
    def __init__(self, workdir, cmd, magnetizingCurves, magnets):
        self.workdir = workdir
        if cmd:
            self.cmd = cmd
        else:
            self.cmd = cfg.get_femag()

        if magnetizingCurves:
            if isinstance(magnetizingCurves,
                          femagtools.mcv.MagnetizingCurve):
                self.magnetizingCurves = magnetizingCurves
            else:
                self.magnetizingCurves = femagtools.mcv.MagnetizingCurve(
                    magnetizingCurves)
        else:
            self.magnetizingCurves = []
        if magnets:
            if isinstance(magnets, femagtools.Magnet):
                self.magnets = magnets
            else:
                self.magnets = femagtools.Magnet(magnets)
        else:
            self.magnets = []

    def copy_magnetizing_curves(self, model, dir=None):
        """extract mc names from model and write files into workdir or dir if given

        Return:
            list of extracted mc names (:obj:`list`)
        """
        dest = dir if dir else self.workdir
        return [self.magnetizingCurves.writefile(m, dest)
                for m in model.set_magcurves(
                        self.magnetizingCurves, self.magnets)]

    def create_fsl(self, pmMachine, operatingConditions):
        """create list of fsl commands"""
        model = femagtools.MachineModel(pmMachine)
        self.modelname = model.name
        self.copy_magnetizing_curves(model)

        builder = femagtools.fsl.Builder()
        return builder.create(model, operatingConditions, self.magnets)

    def get_log_value(self, pattern, modelname='FEMAG-FSL.log'):
        result = []
        pat = re.compile(pattern)
        with open(os.path.join(self.workdir, 'FEMAG-FSL.log')) as f:
            for line in f:
                m = pat.search(line)
                if m:
                    result.append(float(line.split(':')[-1].split()[0]))
        return result

    def read_bch(self, modelname=None):
        "read most recent BCH/BATCH file and return result"
        # read latest bch file if any
        if not modelname:
            modelname = self._get_modelname_from_log()

        result = femagtools.bch.Reader()
        bchfile_list = sorted(glob.glob(os.path.join(
            self.workdir, modelname+'_[0-9][0-9][0-9]'+BCHEXT)))
        if len(bchfile_list) > 0:
            logger.info("Read BCH {}".format(bchfile_list[-1]))
            with io.open(bchfile_list[-1], encoding='latin1',
                         errors='ignore') as f:
                result.read(f)
        return result
    
    def read_los(self, modelname=None):
        "read most recent LOS file and return result"
        # read latest los file if any
        if not modelname:
            modelname = self._get_modelname_from_log()

        losfile_list = sorted(glob.glob(os.path.join(
            self.workdir, modelname+'_[0-9][0-9][0-9].LOS')))
        if len(losfile_list) > 0:
            return ntib.read_los(losfile_list[-1])

        return dict()

    def read_airgap_induction(self):
        """read airgap induction"""
        return ag.read(os.path.join(self.workdir, 'bag.dat'))
    
    def _get_modelname_from_log(self):
        """
        Read the modelname from the Femag Log file
        """
        with open(os.path.join(self.workdir, 'FEMAG-FSL.log')) as f:
            for l in f:
                if l.startswith('New model') or l.startswith('Load model'):
                    model = l.split()[2].replace('"', '').replace(',', '')
                    break
        return model


class Femag(BaseFemag):
    """Invoke and control execution of FEMAG

    Args:
        workdir: name of working directory
        cmd: name of femag program
        magnetizingCurves: collection of lamination material curves
        magnets: collection of magnet material
    """
    def __init__(self, workdir, cmd=None,
                 magnetizingCurves=None, magnets=None):
        super(self.__class__, self).__init__(workdir, cmd,
                                             magnetizingCurves, magnets)

    def run(self, filename, options=['-b'], fsl_args=[]):
        """invoke FEMAG in current workdir

        Args:
            filename: name of file to execute
            options: list of FEMAG options
            fsl_args: list of FSL argument options

        Raises:
            FemagError
        """
        if self.cmd.startswith('wfemag') and \
           '-b' in options and \
           '-m' not in options:
            options.insert(0, '-m')
        args = [self.cmd] + options + [filename] + fsl_args

        basename, ext = os.path.splitext(os.path.basename(filename))
        outname = os.path.join(self.workdir, basename+'.out')
        errname = os.path.join(self.workdir, basename+'.err')
        with open(outname, 'w') as out, open(errname, 'w') as err:
            logger.info('invoking %s', ' '.join(args))
            proc = subprocess.Popen(
                args,
                stdout=out, stderr=err, cwd=self.workdir)

        proc.wait()
        errs = []
        # print femag output
        print(outname)
        with io.open(outname, encoding='latin1', errors='ignore') as outfile:
            for l in outfile:
                print(l.strip())
                if l.find('ERROR') > -1:
                    errs.append(l.strip())

        rc = proc.returncode
        logger.info("%s exited with returncode %d (num errs=%d)",
                    self.cmd, rc, len(errs))
        if rc != 0 or errs:
            with io.open(errname, encoding='latin1',
                         errors='ignore') as errfile:
                for l in errfile:
                    errs.append(l.strip())
            errs.insert(0, 'Exit code {}'.format(rc))
            raise FemagError(errs)
        
    def cleanup(self):
        "removes all created files in workdir"
        if not os.path.exists(self.workdir):
            return
        cleanfiles = ('*.B*CH', '*.I*7-*', '*.A*7-*',
                      '*.dat', '*.LOS', '*.svg', '*.png')
        # '*.TMC','*.TMO', '*.PROT', '*.hxy'):
        for p in cleanfiles:
            for f in glob.glob(os.path.join(self.workdir, p)):
                os.remove(f)

    def __call__(self, pmMachine, operatingConditions,
                 options=['-b'], fsl_args=[]):
        """setup fsl file, run calculation and return BCH results"""
        fslfile = 'femag.fsl'
        with open(os.path.join(self.workdir, fslfile), 'w') as f:
            f.write('\n'.join(self.create_fsl(pmMachine,
                                              operatingConditions)))
        if operatingConditions['calculationMode'] == "pm_sym_loss":
            with open(os.path.join(self.workdir,
                                   self.modelname+'.ntib'), 'w') as f:
                f.write('\n'.join(ntib.create(
                    operatingConditions['speed'],
                    operatingConditions['current'],
                    operatingConditions['angl_i_up'])))
                # TODO: add r1, m

        self.run(fslfile, options, fsl_args)
        if operatingConditions['calculationMode'] == "pm_sym_loss":
            return self.read_los(self.modelname)
        return self.read_bch(self.modelname)


class ZmqFemag(BaseFemag):
    """Invoke and control execution of FEMAG with ZeroMQ

    Args:
        workdir: name of working directory
        cmd: name of femag program
    """
    def __init__(self, workdir, port, host='localhost', cmd=None,
                 magnetizingCurves=None, magnets=None):
        super(self.__class__, self).__init__(workdir, cmd,
                                             magnetizingCurves, magnets)
        self.host = host
        self.port = port
        self.request_socket = None
        self.subscriber_socket = None
        self.proc = None
        self.reader = None

    def __del__(self):
        if self.reader:
            self.reader.continue_loop = False
        logger.debug("Destructor ZmqFemag")

    def __req_socket(self):
        """returns a new request client"""
        if self.request_socket:
            return self.request_socket
        context = zmq.Context.instance()
        self.request_socket = context.socket(zmq.REQ)
        self.request_socket.connect('tcp://{0}:{1}'.format(
            self.host, self.port))
        self.request_socket.setsockopt(zmq.LINGER, 500)
        return self.request_socket

    def __sub_socket(self):
        """returns a subscriber client"""
        if self.subscriber_socket:
            return self.subscriber_socket
        context = zmq.Context.instance()
        self.subscriber_socket = context.socket(zmq.SUB)
        self.subscriber_socket.connect(
            'tcp://{0}:{1}'.format(
                self.host, self.port+1))
        self.subscriber_socket.setsockopt(zmq.SUBSCRIBE, b'')
        self.subscriber_socket.RCVTIMEO = 900  # in milliseconds
        return self.subscriber_socket

    def __is_process_running(self, procId):
        try:
            import psutil
            return psutil.pid_exists(procId)
        except:
            pass
        # long version, self made
        try:
            if procId > 0:
                if platform.system() == "Windows":
                    #                   if procId in psutil.get_pid_list():
                    proc = subprocess.Popen(["tasklist"],
                                            stdout=subprocess.PIPE)
                    for l in proc.stdout:
                        ls = l.split()
                        try:
                            if str(procId) == ls[1]:
                                return True
                        except:
                            continue
                else:
                    if not os.kill(procId, 0):
                        return True
        except OSError as e:
            # No such process
            logger.info("OSError: '{}'\n".format(str(e)))
            return False
        except Exception:
            # we cannot check processId
            logger.info("Error: unknown\n")
            return True
        return False

    def __is_running(self, timeout=1500):
        """check if FEMAG is running in ZMQ mode

        Args:
            timeout: The timeout (in milliseconds) to wait for a response

        Return:
            True if FEMAG is running, False otherwise

        """
        try:
            request_socket = self.__req_socket()
            request_socket.send_string("FSL", flags=zmq.SNDMORE)
            request_socket.send_string("testvar=0")
            poller = zmq.Poller()
            # use POLLIN for recv, POLLOUT for send
            poller.register(request_socket, zmq.POLLIN)
            if poller.poll(timeout):  # ok, femag is running
                logger.info('femag is running for %s',
                            self.request_socket.getsockopt_string(
                                zmq.IDENTITY))
                request_socket.recv_multipart()
                return True
        except Exception as e:
            logger.error(e)
        self.request_socket.close()
        self.request_socket = None
        logger.debug("femag is not running")
        return False

    def send_fsl(self, fsl, callback=None, header='FSL'):
        """sends FSL commands in ZMQ mode and blocks until commands are processed

        Args:
            fsl: string of FSL commands

        Return:
            response
        """
        logger.debug("Send fsl with fsl: {}, callback: {}, header: {}".format(
            fsl, callback, header))

        try:
            # Start the reader thread to get information about the next calculation
            if callback:
                self.reader = FemagReadStream(self.__sub_socket(), callback)
                self.reader.setDaemon(True)
                self.reader.start()

            request_socket = self.__req_socket()
            request_socket.send_string(header, flags=zmq.SNDMORE)
            logger.debug("Sent header")
            request_socket.send_string(fsl)
            logger.debug("Sent fsl wait for response")
            response = request_socket.recv_multipart()
            logger.debug("send_fsl["+fsl+"] done")

            time.sleep(.5)  # Be sure all messages are arrived over zmq
            if callback:
                self.reader.continue_loop = False
            return [s.decode() for s in response]
        except Exception as e:
            logger.exception("send_fsl")
            msg = str(e)
        return ['{"status":"error", "message":"'+msg+'"}', '{}']

    def run(self, options=['-b'], restart=False, procId=None):  # noqa: C901
        """invokes FEMAG in current workdir and returns pid

        Args:
            options: list of FEMAG options

        Raises:
            FemagError
        """
        args = [self.cmd] + options

        if self.__is_running():
            if restart:
                logging.info("must restart")
                self.quit(True)

                # check if process really finished (mq_connection)
                logger.info("procId: %s", procId)
                if procId:
                    for t in range(200):
                        time.sleep(0.1)
                        if not self.__is_process_running(procId):
                            break
                        logger.info(
                            "femag (pid: '{}') not stopped yet".format(
                                procId))
                logger.info("Stopped procId: %s", procId)
            else:
                try:
                    with open(os.path.join(self.workdir,
                                           'femag.pid'), 'r') as pidfile:
                        procId = int(pidfile.readline())
                except Exception:
                    pass
                return procId

        if self.request_socket:
            self.request_socket.close()
        self.request_socket = None

        basename = str(self.port)
        args.append(basename)

        outname = os.path.join(self.workdir, basename+'.out')
        errname = os.path.join(self.workdir, basename+'.err')
        with open(outname, 'w') as out, open(errname, 'w') as err:
            logger.info('invoking %s', ' '.join(args))
            self.proc = subprocess.Popen(
                args,
                stdout=out, stderr=err, cwd=self.workdir)

        # check if mq is ready for listening
        for t in range(200):
            time.sleep(0.1)
            if self.__is_running():
                logger.info("femag (pid: '{}') is listening".format(
                    self.proc.pid))
                break

        # write femag.pid
        logger.info("ready %s", self.proc.pid)
        with open(os.path.join(self.workdir, 'femag.pid'), 'w') as pidfile:
            pidfile.write("{}\n".format(self.proc.pid))
        return self.proc.pid

    def quit(self, save_model=False):
        """terminates femag"""

        if not self.__is_running():
            logger.info("Femag already stopped")
            if self.proc:
                self.proc.wait()
                self.proc = None
            return

        # send exit flags
        self.__req_socket()
        f = '\n'.join(['exit_on_end = true',
                       'exit_on_error = true'])
        response = self.send_fsl(f)

        # send quit command
        try:
            response = self.send_fsl('quit', header='CONTROL')
#                                     recvflags=zmq.NOBLOCK)
        except Exception as e:
            logger.error("Femag Quit zmq message %s", e)

        logger.info("Sent QUIT to femag %s", response)
        # if query, send a answer
        obj = json.loads(response[0])
        logger.debug("status: {}".format(obj['status']))
        if obj['status'] == 'Query':
            logger.info('query: %s => %s',
                        obj['message'], 'saved' if save_model else 'not saved')

            # Only send one msg
            request_socket = self.__req_socket()
            response = request_socket.send_string(
                'Ok' if save_model else 'Cancel')
        
        if self.proc:
            self.proc.wait()
            self.proc = None
        return response

    def stopStreamReader(self):
        if self.reader:
            logger.debug("stop stream reader")
            self.reader.continue_loop = False

    def __call__(self, pmMachine, operatingConditions):
        """setup fsl file, run calculation and return BCH results"""
        self.send_fsl('\n'.join(self.create_fsl(pmMachine,
                                                operatingConditions)))
        return self.read_bch(self.modelname)


class FemagReadStream(Thread):
    def __init__(self, sub_socket, callback):
        Thread.__init__(self)
        logger.debug("Initialize reader thread")
        self.sub_socket = sub_socket
        # TODO: use a function to handle calls without callback
        self.callback = callback if callback else self.dummy_callback
        self.continue_loop = True

    def dummy_callback(self, data):
        """This dummy method is used when no callback is defined.
        """
        logger.warn(
            "No callback defined, fallback to dummy\n >{}".format(data))

    def run(self):
        """Listen for messages from femag as long as the continue_loop is True
        the sub_socket has a timeout to finish the loop and thread after
        the continue_loop is set to False and no messages are arrived.
        """
        logger.debug(
            "Start thread with while condition: {}".format(self.continue_loop))
        while self.continue_loop:
            try:
                response = self.sub_socket.recv_multipart()
                # Sometimes femag send messages with only len = 1. These messages must be ignored
                if len(response) < 2:
                    continue
                # Call the callback function
                self.callback([s.decode() for s in response])
            # The subscriber_socket has a timeout of 900 mil sec. If no answer was arrived
            # this exception is raised - Ingnore
            except zmq.error.Again:
                continue
            # Any other exception is shown in the error log
            except Exception as e:
                logger.error("error in reading output from femag {}".format(e))
                continue
        logger.debug("Exit reader thread")



# -*- coding: utf-8 -*-
"""
    femagtools.airgap
    ~~~~~~~~~~~~~~~~~

    Read airgap dat file



"""
import sys
import numpy as np
import logging

logger = logging.getLogger(__name__)


def read(filename):
    """read dat file with columns (phi, Br, Bphi)
    returns samples, values, amplitude and phase of base harmonic

    Args:
      filename: the ame of the file to be processed
    """
    bag = np.loadtxt(filename).T
    if len(bag) < 3:
        logger.warn("%s has incomplete content", filename)
        return(dict())

    phi = bag[0]
    br = bag[1]
    phisteps = [phi[i+1] - phi[i] for i in range(len(phi)-1)]
    dphi = sum(phisteps)/len(phisteps)
    freq = np.fft.fftfreq(len(br), d=dphi)

    # compute FFT from induction
    Y = 2*np.fft.rfft(br)/len(br)
    
    # find the peak and interpolate
    i = np.argmax(abs(Y))
    logger.debug("Len %d max phi %f max amp %f",
                 len(phi), phi[-1], abs(Y[i]))
    if i == 0 or abs((br[-1] - br[0])/(np.max(br) - np.min(br))) > 1e-2:
        br = np.append(br, -1*br)
        Y = 2*np.fft.rfft(br)/len(br)
        freq = np.fft.fftfreq(len(br), d=dphi)
        i = np.argmax(abs(Y))

    if freq[i] > 0:
        logger.info("%s: Period %f max phi %f B amp %f",
                    filename, 1/freq[i], len(br)*dphi, abs(Y[i]))
        a = abs(Y[i])
        T0 = abs(1/freq[i])
        x0 = np.angle(Y[i])
        phi = [ix*dphi for ix in range(len(br))]
        x = np.array([2*x/T0*np.pi for x in phi])

    return dict(Bamp=a,
                phi0=x0,
                pos=phi,
                B=br.tolist(),
                B_fft=(a*np.cos(x+x0)).tolist())

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    else:
        filename = sys.stdin.readline().strip()
            
    print(read(filename)['Bamp'])



# -*- coding: utf-8 -*-
"""
    femagtools.grid
    ~~~~~~~~~~~~~~~

    Parameter range calculation



"""
import logging
import glob
import os
import numpy as np
import femagtools
import femagtools.model
import femagtools.fsl
import femagtools.condor
import femagtools.moproblem

logger = logging.getLogger(__name__)


def baskets(items, basketsize=10):
    """generates balanced baskets from iterable, contiguous items"""
    num_items = len(items)
    num_baskets = max(1, num_items//basketsize)
    if num_items % basketsize and basketsize < num_items:
        num_baskets += 1
    step = num_items//num_baskets
    for i in range(0, num_items, step):
        yield items[i:i+step]


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i+n]


def create_parameter_range(domain):
    """returns the transposed array of the combined domain values"""
    L = [len(d) for d in domain]
    LS = np.prod(L)
    s = []
    e = 1
    for d in domain:
        LS = LS//len(d)
        s.append(np.repeat(d*LS, e))
        e = e*L[0]
        L = L[1:]
    return np.array(s).T


class Grid(object):
    """Parameter variation calculation"""
    def __init__(self, workdir,
                 magnetizingCurves=None, magnets=None):
        self.femag = femagtools.Femag(workdir,
                                      magnetizingCurves=magnetizingCurves,
                                      magnets=magnets)
        self.stop = False  # rudimentary: gives the ability to stop a running parameter variation. thomas.maier/OSWALD
        """
        the "owner" of the Grid have to take care to terminate all running xfemag64 or wfemagw64
        processes after setting stop to True
        For example:
        
        def killFemagThreads():
            if sys.platform.startswith('linux'):
                os.system("kill $(ps aux | grep '[x]femag64' | awk '{print $2}')")
            else:
                os.system("taskkill /f /im wfemagw64.exe")
                
        thomas.maier/OSWALD
        """

    def setup_model(self, builder, model):
        """builds model in current workdir and returns its filenames"""
        # get and write mag curves
        mc_files = self.femag.copy_magnetizing_curves(model)
        
        filename = 'femag.fsl'
        logger.info("setup model in %s", self.femag.workdir)
        with open(os.path.join(self.femag.workdir, filename), 'w') as f:
            f.write('\n'.join(builder.create_model(model,
                                                   self.femag.magnets) +
                              ['save_model(close)']))

        self.femag.run(filename, options=['-b'])
        model_files = [os.path.join(self.femag.workdir, m)
                       for m in mc_files] + \
                           glob.glob(os.path.join(self.femag.workdir,
                                                  model.name+'_*.poc')) + \
                            glob.glob(os.path.join(self.femag.workdir,
                                                   model.name+'*7'))
        
        logger.info("model %s created", model.name)
        return model_files
    
    def __call__(self, opt, pmMachine, operatingConditions,
                 engine, bchMapper=None):
        """calculate objective vars for all decision vars"""

        self.stop = False  # make sure the calculation will start. thomas.maier/OSWALD

        decision_vars = opt['decision_vars']
        objective_vars = opt.get('objective_vars', {})

        steps = [d.get('steps', 10) for d in decision_vars]
        logger.info('STEPS %s', str(steps))

        model = femagtools.model.MachineModel(pmMachine)
        # check if this model needs to be modified
        immutable_model = len([d for d in decision_vars
                               if hasattr(model,
                                          d['name'].split('.')[0])]) == 0
        operatingConditions['lfe'] = model.lfe
        operatingConditions['move_action'] = model.move_action
        operatingConditions['pocfilename'] = (model.get('name') +
                                              '_' + str(model.get('poles')) +
                                              'p.poc')
        operatingConditions.update(model.windings)
        fea = femagtools.model.FeaModel(operatingConditions)

        prob = femagtools.moproblem.FemagMoProblem(decision_vars,
                                                   objective_vars)

        job = engine.create_job(self.femag.workdir)
        builder = femagtools.fsl.Builder()

        # build x value array
        
        domain = [list(np.linspace(l, u, s))
                  for s, l, u in zip(steps, prob.lower, prob.upper)]

        par_range = create_parameter_range(domain)
        f = []
        p = 1
        logger.debug(par_range)

        if immutable_model:
            modelfiles = self.setup_model(builder, model)
            logger.info("Files %s", modelfiles)

        self.bchmapper_data = []  # clear bch data
        # split x value (par_range) array in handy chunks:
        for population in baskets(par_range, opt['population_size']):
            if self.stop:  # try to return the results so far. thomas.maier/OSWALD
                logger.info('stopping grid execution... returning results so far...')
                try:
                    shape = [len(objective_vars)] + [len(d) for d in reversed(domain)]
                    logger.debug("BEFORE: f shape %s --> %s", np.shape(np.array(f).T), shape)
                    complete = int(reduce((lambda x, y: x * y), [len(z) for z in domain]))
                    logger.debug("need {} in total".format(complete))
                    remaining = complete - int(np.shape(np.array(f).T)[1])
                    values = int(np.shape(np.array(f).T)[0])
                    logger.debug("going to append {} None values".format(remaining))
                    f += remaining * [values * [np.nan]]
                    shape = [len(objective_vars)] + [len(d) for d in reversed(domain)]
                    logger.debug("AFTER: f shape %s --> %s", np.shape(np.array(f).T), shape)
                    objectives = np.reshape(np.array(f).T, shape)
                    r = dict(f=objectives.tolist(),
                             x=domain)
                    return r
                except:
                    return {}
                    pass

            logger.info('........ %d / %d', p, len(par_range)//len(population))
            job.cleanup()
            for k, x in enumerate(population):
                task = job.add_task()
                if immutable_model:
                    prob.prepare(x, fea)
                    for m in modelfiles:
                        task.add_file(m)
                    task.add_file('femag.fsl',
                                  builder.create_open(model) +
                                  builder.create_common(model) +
                                  builder.create_analysis(fea))
                else:
                    try:
                        prob.prepare(x, model)
                    except:
                        prob.prepare(x, [model, fea])
                    logger.info("prepare %s", x)
                    for mc in self.femag.copy_magnetizing_curves(
                            model,
                            task.directory):
                        task.add_file(mc)

                    task.add_file('femag.fsl',
                                  builder.create_model(model,
                                                       self.femag.magnets) +
                                  builder.create_analysis(fea))

            status = engine.submit()
            logger.info('Started %s', status)
            if bchMapper and isinstance(engine, femagtools.condor.Engine):
                 return {}  # BatchCalc Mode
            status = engine.join()

            for t in job.tasks:
                if t.status == 'C':
                    r = t.get_results()
                    if bchMapper:  # Mode => collectBchData
                        self.addBchMapperData(bchMapper(r))
                    if isinstance(r, dict) and 'error' in r:
                        logger.warn("job %d failed: %s", k, r['error'])
                        f.append([float('nan')]*len(objective_vars))
                    else:
                        prob.setResult(bchMapper(r)) if bchMapper else prob.setResult(r)

                        f.append(prob.objfun([]))
            p += 1

        logger.info('...... DONE')
        logger.debug("Result %s", np.shape(f))

        shape = [len(objective_vars)] + [len(d) for d in reversed(domain)]
        logger.info("f shape %s --> %s", np.shape(np.array(f).T), shape)
        objectives = np.reshape(np.array(f).T, shape)
        return dict(f=objectives.tolist(),
                    x=domain)

    def addBchMapperData(self, bchData):
        self.bchmapper_data.append(bchData)

    def getBchMapperData(self):
        return self.bchmapper_data



# -*- coding: utf-8 -*-
"""
    femagtools.isa7
    ~~~~~~~~~~~~~~~

    Read FEMAG I7/ISA7 model files
"""
import logging
import struct
import sys
import itertools
import re
import numpy as np
from collections import Counter, defaultdict

logger = logging.getLogger('femagtools.isa7')


class Reader(object):
    """
    Open and Read I7/ISA7 file

    Arguments:
        filename: name of I7/ISA7 file to be read
    """

    def __init__(self, filename):
        with open(filename, mode="rb") as self.file:
            self.file = self.file.read()
        self.pos = 0

    def next_block(self, fmt):
        """
        Read binary data and return unpacked values according to format string.

        Arguments:
            fmt: Format string (see python struct module)
        """

        fmt_ = fmt.replace("?", "i")

        blockSize = struct.unpack_from("=i", self.file, self.pos)[0]
        self.pos += 4
        try:
            unpacked = struct.iter_unpack("=" + fmt_,
                                          self.file[self.pos:self.pos
                                                    + blockSize])
            unpacked = [x for x in unpacked]

        except AttributeError:  # python 2 has no iter_unpack
            chunksize = struct.calcsize("=" + fmt_)
            offset = self.pos
            unpacked = []
            for j in range(blockSize // chunksize):
                unpacked.append(struct.unpack_from("=" + fmt_,
                                                   self.file,
                                                   offset))
                offset += chunksize
            logger.info("%s: %d %d", fmt_, blockSize, len(unpacked))

        self.pos += blockSize + 4

        fmt_ = ""
        for s in re.findall("[0-9]*.|[0-9]*\?", fmt):
            if len(s) > 1 and s[-1] != "s":
                fmt_ += int(s[:-1]) * s[-1]
            else:
                fmt_ += s
        values = []
        for i, dtype in enumerate(re.findall("\?|[0-9]*s?", fmt_)[:-1]):
            if dtype == "?":
                values.append([bool(u[i]) for u in unpacked])
            elif "s" in dtype:
                values.append([u[i].decode() for u in unpacked])
            else:
                values.append([u[i] for u in unpacked])

        if len(fmt) == 1:
            return values[0]
        else:
            return values

    def skip_block(self, skips=1):
        """
        Proceed to the next block without reading any data.

        Arguments:
            skips: number of blocks to be skipped
        """
        while skips > 0:
            blockSize = struct.unpack_from("=i", self.file, self.pos)[0]
            self.pos += 4 + blockSize + 4
            skips -= 1


class Isa7(object):
    """
    The ISA7 Femag model

    Arguments:
        filename: name of I7/ISA7 file to be read
    """

    color = {1: [1.0, 0.0, 0.0],
             2: [0.0, 1.0, 0.0],
             3: [1.0, 1.0, 0.0],
             4: [0.0, 0.5019607843137255, 1.0],
             5: [0.9803921568627451, 0.0, 1.0],
             6: [0.0, 1.0, 0.8235294117647058],
             7: [1.0, 1.0, 1.0],
             8: [0.0, 0.0, 0.0],
             9: [0.0, 0.0, 0.5882352941176471],
             10: [0.6666666666666666, 0.0, 0.0],
             11: [0.6666666666666666, 1.0, 0.0],
             12: [1.0, 0.6274509803921569, 0.0],
             13: [0.0, 0.0, 1.0],
             14: [0.6666666666666666, 0.0, 1.0],
             15: [0.0, 0.8235294117647058, 1.0],
             16: [0.8274509803921568, 0.8274509803921568, 0.8274509803921568]}

    def __init__(self, filename):
        self.__read(filename)

        self.points = []
        for p in range(self.NUM_PNT):
            self.points.append(
                Point(self.POINT_ISA_PT_VALID,
                      self.POINT_ISA_POINT_REC_PT_CO_X[p],
                      self.POINT_ISA_POINT_REC_PT_CO_Y[p]))

        self.lines = []
        for ln in range(self.NUM_LIN):
            pk1 = self.LINE_ISA_LINE_REC_LN_PNT_1[ln]
            pk2 = self.LINE_ISA_LINE_REC_LN_PNT_2[ln]

            point1 = self.points[abs(pk1) - 1]
            point2 = self.points[abs(pk2) - 1]

            if pk1 > 0 and pk2 > 0:
                self.lines.append(
                    Line(self.LINE_ISA_LN_VALID, point1, point2))
            else:
                self.lines.append(
                    Line(self.LINE_ISA_LN_VALID, point1, point2))

        self.nodes = []
        for n in range(self.NUM_NOD):
            self.nodes.append(
                Node(self.NODE_ISA_ND_VALID[n],
                     n + 1,
                     self.NODE_ISA_NODE_REC_ND_BND_CND[n],
                     self.NODE_ISA_NODE_REC_ND_PER_NOD[n],
                     self.NODE_ISA_NODE_REC_ND_CO_1[n],
                     self.NODE_ISA_NODE_REC_ND_CO_2[n],
                     self.NODE_ISA_NODE_REC_ND_VP_RE[n],
                     self.NODE_ISA_NODE_REC_ND_VP_IM[n]))

        self.nodechains = []
        for nc in range(self.NUM_NDCH):
            nd1 = self.NDCHN_ISA_NDCHN_REC_NC_NOD_1[nc]
            nd2 = self.NDCHN_ISA_NDCHN_REC_NC_NOD_2[nc]
            ndm = self.NDCHN_ISA_NDCHN_REC_NC_NOD_MID[nc]

            node1 = self.nodes[abs(nd1) - 1]
            nodem = self.nodes[ndm - 1]
            node2 = self.nodes[abs(nd2) - 1]

            if nd1 < 0 or nd2 < 0:
                nodes = node1, nodem, node2
            elif ndm > 0:
                nodes = node1, nodem, node2
            else:
                nodes = node1, None, node2

            self.nodechains.append(
                NodeChain(self.NDCHN_ISA_NC_VALID, nc + 1, nodes))

        self.elements = []
        for e in range(self.NUM_ELE):
            ndkeys = []
            ndk = self.ELEM_ISA_EL_NOD_PNTR[e]

            while ndk > 0:
                ndkeys.append(self.ELE_NOD_ISA_ND_KEY[ndk - 1])
                ndk = self.ELE_NOD_ISA_NXT_ND_PNTR[ndk - 1]

            vertices = [self.nodes[k - 1] for k in ndkeys]

            self.elements.append(
                Element(self.ELEM_ISA_EL_VALID[e],
                        e + 1,
                        self.ELEM_ISA_ELEM_REC_EL_TYP[e],
                        self.ELEM_ISA_ELEM_REC_EL_SE_KEY[e] - 1,
                        vertices,
                        (self.ELEM_ISA_ELEM_REC_EL_RELUC[e],
                         self.ELEM_ISA_ELEM_REC_EL_RELUC_2[e]),
                        (self.ELEM_ISA_ELEM_REC_EL_MAG_1[e],
                         self.ELEM_ISA_ELEM_REC_EL_MAG_2[e])))

        self.superelements = []
        for se in range(self.NUM_SPEL):
            nc_keys = []
            nc_ptr = self.SUPEL_ISA_SE_NDCHN_PNTR[se]

            while nc_ptr > 0:
                nc_keys.append(self.SE_NDCHN_ISA_NC_KEY[nc_ptr - 1])
                nc_ptr = self.SE_NDCHN_ISA_NXT_NC_PNTR[nc_ptr - 1]

            nodechains = []
            for nck in nc_keys:
                if nck > 0:
                    nodechains.append(self.nodechains[abs(nck) - 1])
                else:
                    nodechains.append(self.nodechains[abs(nck) - 1].reverse())

            el_keys = []
            el_ptr = self.SUPEL_ISA_SE_EL_PNTR[se]

            while el_ptr > 0:
                el_keys.append(self.SE_EL_ISA_EL_KEY[el_ptr - 1])
                el_ptr = self.SE_EL_ISA_NXT_EL_PNTR[el_ptr - 1]

            elements = []
            for elk in el_keys:
                elements.append(self.elements[elk - 1])

            self.superelements.append(
                SuperElement(self.SUPEL_ISA_SE_VALID[se],
                             se + 1,
                             self.SUPEL_ISA_SUPEL_REC_SE_SR_KEY[se] - 1,
                             elements,
                             nodechains,
                             self.SUPEL_ISA_SUPEL_REC_SE_COL[se],
                             nc_keys,
                             self.SUPEL_ISA_SUPEL_REC_SE_MCV_TYP,
                             self.SUPEL_ISA_SUPEL_REC_SE_COND_TYP,
                             self.SUPEL_ISA_SUPEL_REC_SE_CONDUC,
                             self.SUPEL_ISA_SUPEL_REC_SE_LENGHT,
                             self.SUPEL_ISA_SUPEL_REC_SE_VEL_SYS,
                             self.SUPEL_ISA_SUPEL_REC_SE_VELO_1,
                             self.SUPEL_ISA_SUPEL_REC_SE_VELO_2,
                             self.SUPEL_ISA_SUPEL_REC_SE_CURD_RE,
                             self.SUPEL_ISA_SUPEL_REC_SE_CURD_IM))

        self.subregions = []
        for sr in range(self.NUM_SR):
            se_keys = []
            se_ptr = self.SR_ISA_SR_SE_PNTR[sr]

            while se_ptr > 0:
                se_keys.append(self.SR_SE_ISA_SE_KEY[se_ptr - 1])
                se_ptr = self.SR_SE_ISA_NXT_SE_PNTR[se_ptr - 1]

            superelements = []
            for sek in se_keys:
                superelements.append(self.superelements[sek - 1])

            nodechains = []
            nc_keys = []
            for se in superelements:
                nc_keys.extend([abs(nc.key) for nc in se.nodechains])
            nc_keys = [nck for nck, count
                       in Counter(nc_keys).items() if count < 2]
            for se in superelements:
                nodechains.extend([nc
                                   for nc in se.nodechains
                                   if abs(nc.key) in nc_keys])
                
            self.subregions.append(
                SubRegion(self.SR_ISA_SR_VALID[sr],
                          sr + 1,
                          self.SR_ISA_SR_REC_SR_TYP[sr],
                          self.SR_ISA_SR_REC_SR_COL[sr],
                          self.SR_ISA_SR_REC_SR_NAME[sr],
                          self.SR_ISA_SR_REC_SR_CUR_DIR[sr],
                          self.SR_ISA_SR_REC_SR_WB_KEY[sr] - 1,
                          superelements,
                          nodechains))

        self.windings = []
        for wd in range(self.NUM_WB):
            sr_keys = []
            sr_ptr = self.WB_ISA_WB_SR_PNTR[wd]

            while sr_ptr > 0:
                sr_keys.append(self.WB_SR_ISA_SR_KEY[sr_ptr - 1])
                sr_ptr = self.WB_SR_ISA_NXT_SR_PNTR[sr_ptr - 1]

            subregions = []
            for srk in sr_keys:
                subregions.append(self.subregions[srk - 1])

            self.windings.append(
                Winding(self.WB_ISA_WB_VALID[wd],
                        wd + 1,
                        self.WB_ISA_WB_REC_WB_NAME[wd],
                        subregions,
                        self.WB_ISA_WB_REC_WB_TURN[wd],
                        self.WB_ISA_WB_REC_WB_GCUR_RE[wd],
                        self.WB_ISA_WB_REC_WB_GCUR_IM[wd],
                        self.WB_ISA_WB_REC_WB_IMPDZ_RE[wd],
                        self.WB_ISA_WB_REC_WB_IMPDZ_IM[wd],
                        self.WB_ISA_WB_REC_WB_VOLT_RE[wd],
                        self.WB_ISA_WB_REC_WB_VOLT_IM[wd]))

    def __read(self, filename):
        reader = Reader(filename)

        (self.NUM_PNT, self.PNT_PTR, self.PNT_HIDX,
         self.NUM_LIN, self.LIN_PTR, self.LIN_HIDX,
         self.NUM_NOD, self.NOD_PTR, self.NOD_HIDX,
         self.NUM_NDEL, self.NDEL_PTR, self.NDEL_HIDX,
         self.NUM_NDCH, self.NDCH_PTR, self.NDCH_HIDX,
         self.NUM_ELE, self.ELE_PTR, self.ELE_HIDX,
         self.NUM_ELND, self.ELND_PTR, self.ELND_HIDX,
         self.NUM_SPEL, self.SPEL_PTR, self.SPEL_HIDX,
         self.NUM_SE_EL, self.SE_EL_PTR, self.SE_EL_HIDX,
         self.NUM_SPEL_NDCH, self.SPEL_NDCH_PTR, self.SPEL_NDCH_HIDX,
         self.NUM_SR, self.SR_PTR, self.SR_HIDX,
         self.NUM_SR_SE, self.SR_SE_PTR, self.SR_SE_HIDX,
         self.NUM_WB, self.WB_PTR, self.WB_HIDX,
         self.NUM_WB_SR, self.WB_SR_PTR, self.WB_SR_HIDX,
         self.NUM_OB, self.OB_PTR, self.OB_HIDX,
         self.NUM_OB_SR, self.OB_SR_PTR, self.OB_SR_HIDX,
         self.NUM_DV, self.DV_PTR, self.DV_HIDX,
         self.NUM_DV_OB, self.DV_OB_PTR, self.DV_OB_HIDX,
         self.NUM_MC, self.MC_PTR, self.MC_HIDX,
         self.NUM_CF, self.CF_PTR, self.CF_HIDX,
         self.NUM_CF_MC, self.CF_MC_PTR, self.CF_MC_HIDX,
         self.NUM_WN, self.WN_PTR, self.WN_HIDX,
         self.NUM_WN_SW, self.WN_SW_PTR, self.WN_SW_HIDX
         ) = reader.next_block("i")

        (self.POINT_ISA_PT_VALID,
         self.POINT_ISA_POINT_REC_PT_CO_X,
         self.POINT_ISA_POINT_REC_PT_CO_Y) = reader.next_block("?ff")

        (self.LINE_ISA_LN_VALID,
         self.LINE_ISA_LINE_REC_LN_PNT_1,
         self.LINE_ISA_LINE_REC_LN_PNT_2) = reader.next_block("?hh")

        (self.NODE_ISA_ND_VALID,
         self.NODE_ISA_NOD_EL_PNTR,
         self.NODE_ISA_ND_CO_RAD,
         self.NODE_ISA_ND_CO_PHI,
         self.NODE_ISA_NODE_REC_ND_BND_CND,
         self.NODE_ISA_NODE_REC_ND_PER_NOD,
         self.NODE_ISA_NODE_REC_ND_SV_PNTR,
         self.NODE_ISA_NODE_REC_ND_CO_1,
         self.NODE_ISA_NODE_REC_ND_CO_2,
         self.NODE_ISA_NODE_REC_ND_VP_RE,
         self.NODE_ISA_NODE_REC_ND_VP_IM) = reader.next_block("?iffhiiffff")

        (self.NOD_ELE_ISA_EL_KEY,
         self.NOD_ELE_ISA_NXT_EL_PNTR) = reader.next_block("ii")

        (self.NDCHN_ISA_NC_VALID,
         self.NDCHN_ISA_NDCHN_REC_NC_NOD_1,
         self.NDCHN_ISA_NDCHN_REC_NC_NOD_2,
         self.NDCHN_ISA_NDCHN_REC_NC_NOD_MID) = reader.next_block("?iii")

        (self.ELEM_ISA_EL_VALID,
         self.ELEM_ISA_EL_NOD_PNTR,
         self.ELEM_ISA_ELEM_REC_EL_TYP,
         self.ELEM_ISA_ELEM_REC_EL_SE_KEY,
         self.ELEM_ISA_ELEM_REC_EL_RELUC,
         self.ELEM_ISA_ELEM_REC_EL_RELUC_2,
         self.ELEM_ISA_ELEM_REC_EL_MAG_1,
         self.ELEM_ISA_ELEM_REC_EL_MAG_2) = reader.next_block("?ihhffff")

        (self.ELE_NOD_ISA_ND_KEY,
         self.ELE_NOD_ISA_NXT_ND_PNTR) = reader.next_block("ii")

        (self.SUPEL_ISA_SE_VALID,
         self.SUPEL_ISA_SE_NDCHN_PNTR,
         self.SUPEL_ISA_SE_EL_PNTR,
         self.SUPEL_ISA_SUPEL_REC_SE_COL,
         self.SUPEL_ISA_SUPEL_REC_SE_MCV_TYP,
         self.SUPEL_ISA_SUPEL_REC_SE_COND_TYP,
         self.SUPEL_ISA_SUPEL_REC_SE_VEL_SYS,
         self.SUPEL_ISA_SUPEL_REC_SE_SR_KEY,
         self.SUPEL_ISA_SUPEL_REC_SE_VELO_1,
         self.SUPEL_ISA_SUPEL_REC_SE_VELO_2,
         self.SUPEL_ISA_SUPEL_REC_SE_CONDUC,
         self.SUPEL_ISA_SUPEL_REC_SE_LENGHT,
         self.SUPEL_ISA_SUPEL_REC_SE_CURD_RE,
         self.SUPEL_ISA_SUPEL_REC_SE_CURD_IM
         ) = reader.next_block("?iihhhhhffffff")

        (self.SE_NDCHN_ISA_NC_KEY,
         self.SE_NDCHN_ISA_NXT_NC_PNTR) = reader.next_block("ii")

        (self.SE_EL_ISA_EL_KEY,
         self.SE_EL_ISA_NXT_EL_PNTR) = reader.next_block("ii")

        (self.SR_ISA_SR_VALID,
         self.SR_ISA_SR_SE_PNTR,
         self.SR_ISA_SR_REC_SR_TYP,
         self.SR_ISA_SR_REC_SR_COL,
         self.SR_ISA_SR_REC_SR_NAME,
         self.SR_ISA_SR_REC_SR_CUR_DIR,
         self.SR_ISA_SR_REC_SR_WB_KEY,
         self.SR_ISA_SR_REC_SR_NTURNS,
         self.SR_ISA_SR_REC_SR_SV_PNTR,
         self.SR_ISA_SR_REC_SR_ARRAY,
         self.SR_ISA_SR_REC_SR_GCUR_RE,
         self.SR_ISA_SR_REC_SR_GCUR_IM,
         self.SR_ISA_SR_REC_SR_VOLT_RE,
         self.SR_ISA_SR_REC_SR_VOLT_IM) = reader.next_block("?hhh4shhhhfffff")

        (self.SR_SE_ISA_SE_KEY,
         self.SR_SE_ISA_NXT_SE_PNTR) = reader.next_block("hh")

        (self.WB_ISA_WB_VALID,
         self.WB_ISA_WB_SR_PNTR,
         self.WB_ISA_WB_REC_WB_COL,
         self.WB_ISA_WB_REC_WB_NAME,
         self.WB_TURN,
         self.WB_ISA_WB_REC_WB_SR_NUM,
         self.WB_ISA_WB_REC_WB_WND_KEY,
         self.WB_ISA_WB_REC_WB_UNIT_RES,
         self.WB_ISA_WB_REC_WB_GCUR_RE,
         self.WB_ISA_WB_REC_WB_GCUR_IM,
         self.WB_ISA_WB_REC_WB_VOLT_RE,
         self.WB_ISA_WB_REC_WB_VOLT_IM,
         self.WB_ISA_WB_REC_WB_IMPDZ_RE,
         self.WB_ISA_WB_REC_WB_IMPDZ_IM) = reader.next_block("?hh4shhhfffffff")

        self.WB_ISA_WB_REC_WB_TURN = []
        for wd in range(self.NUM_WB):
            if self.WB_ISA_WB_REC_WB_UNIT_RES[wd] == 0:
                self.WB_ISA_WB_REC_WB_TURN.append(
                    self.WB_TURN[wd])
            else:
                self.WB_ISA_WB_REC_WB_TURN.append(
                    self.WB_ISA_WB_REC_WB_UNIT_RES[wd])
                self.WB_ISA_WB_REC_WB_UNIT_RES[wd] = 0

        (self.WB_SR_ISA_SR_KEY,
         self.WB_SR_ISA_NXT_SR_PNTR) = reader.next_block("hh")

        reader.skip_block(21)

        self.ANZAHL_TG = reader.next_block("iii")[1][0]

        reader.skip_block(7)
        reader.skip_block(self.ANZAHL_TG + 1)
        reader.skip_block(1)

        self.FC_RADIUS = reader.next_block("f")[0]

    def to_gmsh(self, fileformat="geo", extrude=0, layers=0, recombine=False):
        """
        Convert model to gmsh file
        
        Arguments:
            fileformat: ['geo' | 'msh']
            extrude: extrude surfaces using a translation along the z-axis
            layers:
            recombine:
        """
        el_number = itertools.count(1)

        airgap_center_elements = []
        for e in self.elements:
            outside = [np.sqrt(v.x**2 + v.y**2) > self.FC_RADIUS
                       for v in e.vertices]
            if any(outside) and not all(outside):
                airgap_center_elements.append(e)

        airgap_center_vertices = [v for e in airgap_center_elements
                                  for v in e.vertices]

        airgap_rotor_elements = []
        airgap_stator_elements = []
        for e in self.elements:
            if e in airgap_center_elements:
                continue
            for v in e.vertices:
                if v not in airgap_center_vertices:
                    continue
                if np.sqrt(v.x**2 + v.y**2) > self.FC_RADIUS:
                    airgap_stator_elements.append(e)
                    break
                else:
                    airgap_rotor_elements.append(e)
                    break

        airgap_stator_vertices = [v for e in airgap_stator_elements
                                  for v in e.vertices]

        airgap_lines = []
        for e in airgap_center_elements:
            ev = e.vertices
            for i, v1 in enumerate(ev):
                v2 = ev[i-1]
                if v1 in airgap_stator_vertices and \
                   v2 in airgap_stator_vertices:
                    airgap_lines.append((v1, v2))

        nodechain_links = defaultdict(lambda: [])
        for nc in self.nodechains:
            nodechain_links[nc.node1].extend(nc.nodes)
            nodechain_links[nc.node2].extend(nc.nodes)
            if nc.nodemid is not None:
                nodechain_links[nc.nodemid].extend(nc.nodes)

        physical_lines = ["v potential 0",
                          "v potential const",
                          "periodic +",
                          "periodic -",
                          "infinite boundary",
                          "no condition",
                          "Airgap"]

        physical_surfaces = sorted(set([sr.name
                                        for sr in self.subregions]
                                       + ["Winding {} {}".format(w.key, pol)
                                          for w in self.windings
                                          for pol in ("+", "-")]
                                       + ["Air Rotor",
                                          "Air Stator",
                                          "Airgap Rotor",
                                          "Airgap Stator",
                                          "PM1", "PM2",
                                          "PM3", "PM4"]))

        def physical_line(n1, n2):
            if (n1, n2) in airgap_lines or (n2, n1) in airgap_lines:
                return 7  # airgap
            if n1.bndcnd == n2.bndcnd:
                return boundary_condition(n1)
            if boundary_condition(n1) == 1:
                return boundary_condition(n2)
            return boundary_condition(n1)
        
        def boundary_condition(node):
            if node.bndcnd == 0:
                return 6  # no condition
            if node.bndcnd == 1:
                return 1  # vpot 0
            if node.bndcnd == 2:
                return 2  # vpot const
            if node.bndcnd == 3 or node.bndcnd == 6:
                return 4  # periodic -
            if node.bndcnd == 4 or node.bndcnd == 5:
                return 3  # periodic +
            if node.bndcnd == 8 or node.bndcnd == 9:
                return 1  # vpot 0
            
        def physical_surface(e):

            def surface_id(name):
                return physical_surfaces.index(name) + len(physical_lines) + 1
            
            if any(e.mag):
                if e.mag[0] > 0:
                    if e.mag[1] > 0:
                        return surface_id("PM1")
                    return surface_id("PM2")
                if e.mag[1] > 0:
                    return surface_id("PM3")
                return surface_id("PM4")
            
            if e in airgap_rotor_elements or e in airgap_center_elements:
                return surface_id("Airgap Rotor")

            if e in airgap_stator_elements:
                return surface_id("Airgap Stator")
            
            sr_key = self.superelements[e.se_key].sr_key
            if sr_key == -1:
                v = e.vertices[0]
                if np.sqrt(v.x**2 + v.y**2) > self.FC_RADIUS:
                    return surface_id("Air Stator")
                return surface_id("Air Rotor")

            sr = self.subregions[sr_key]
            if sr.wb_key != -1:
                wb = self.subregions[sr.wb_key]
                if sr.curdir > 0:
                    return surface_id("Winding {} -".format(wb.key)) 
                return surface_id("Winding {} +".format(wb.key))
            
            return surface_id(sr.name)

        def line_on_boundary(n1, n2):
            if n1.on_boundary() and n2.on_boundary():
                if n1 in nodechain_links.keys():
                    return n2 in nodechain_links[n1]
                else:
                    return False
            else:
                return (n1, n2) in airgap_lines or (n2, n1) in airgap_lines

        def get_msh():
            msh = ["$MeshFormat",
                   "2.2 0 8",
                   "$EndMeshFormat"]

            msh.append("$PhysicalNames")
            msh.append("{}".format(len(physical_lines)
                                   + len(physical_surfaces)))
            for i, n in enumerate(physical_lines):
                msh.append("1 {} \"{}\"".format(i+1, n))
            for i, n in enumerate(physical_surfaces):
                msh.append("2 {} \"{}\"".format(i + 1 + len(physical_lines), n))
            msh.append("$EndPhysicalNames")

            msh.append("$Nodes")
            msh.append("{}".format(len(self.nodes)))
            for n in self.nodes:
                msh.append("{} {} {} 0".format(n.key, n.x, n.y))
            msh.append("$EndNodes")

            msh.append("$Elements")
            msh.append("{}".format(len(self.elements)
                                   # vertices on boundary
                                   + len([n for n in self.nodes
                                          if n.on_boundary() and
                                          n in nodechain_links.keys()])
                                   + len(airgap_lines)))

            for e in self.elements:
                ev = e.vertices

                # 2-node lines
                for i, v in enumerate(ev):
                    v1 = v
                    v2 = ev[i-1]
                    if line_on_boundary(v1, v2):
                        msh.append(" ".join(map(str, [
                            next(el_number),
                            1,
                            2,
                            physical_line(v1, v2),
                            physical_line(v1, v2),
                            v2.key,
                            v1.key])))

                # 3-node triangles
                if len(ev) == 3:
                    msh.append(" ".join(map(str, [
                        next(el_number),
                        2,
                        2,
                        physical_surface(e),
                        physical_surface(e) + len(self.nodes),
                        ev[0].key,
                        ev[1].key,
                        ev[2].key])))

                # 4-node quadrangles
                elif len(ev) == 4:
                    msh.append(" ".join(map(str, [
                        next(el_number),
                        3,
                        2,
                        physical_surface(e),
                        physical_surface(e) + len(self.nodes),
                        ev[0].key,
                        ev[1].key,
                        ev[2].key,
                        ev[3].key])))

                # 6-node second order triangles
                elif len(ev) == 6:
                    msh.append(" ".join(map(str, [
                        next(el_number),
                        9,
                        2,
                        physical_surface(e),
                        physical_surface(e) + len(self.nodes),
                        ev[0].key,
                        ev[1].key,
                        ev[2].key,
                        ev[3].key,
                        ev[4].key,
                        ev[5].key])))

                # 8-node second order quadrangles
                elif len(ev) == 8:
                    msh.append(" ".join(map(str, [
                        next(el_number),
                        16,
                        2,
                        physical_surface(e),
                        physical_surface(e) + len(self.nodes),
                        ev[0].key,
                        ev[1].key,
                        ev[2].key,
                        ev[3].key,
                        ev[4].key,
                        ev[5].key,
                        ev[6].key,
                        ev[7].key])))

                else:
                    logger.warn("element {0} has an unsupported \
                                 number of nodes ({1})".format(e.key, len(ev)))

            msh.append("$EndElements")
            return msh

        def get_geo():
            geo = []
            nc_nodes = set([n for nc in self.nodechains for n in nc.nodes])

            for n in self.nodes:
                if n in nc_nodes:
                    geo.append("Point({}) = {{{}, {}, {}}};".format(
                        n.key, n.x, n.y, 0))
                    
            for nc in self.nodechains:
                geo.append("Line({}) = {{{}}};".format(
                        nc.key, ", ".join([str(n.key) for n in nc.nodes])))
            used = set()
            for nc in self.nodechains:
                n1, n2 = nc.nodes[0], nc.nodes[1]
                if line_on_boundary(n1, n2):
                    id_ = physical_line(n1, n2)
                    name = physical_lines[id_ - 1]
                    if extrude:
                        geo.append(
                            "extrusion[] = Extrude {{0, 0, {}}} {{ Line{{{}}}; {}{}}};".format(
                                extrude,
                                nc.key,
                                "Layers{{{}}}; ".format(layers) if layers else "",
                                "Recombine; " if recombine else ""))
                        
                        geo.append("Physical Surface('{}', {}) {} extrusion[1];".format(
                            name,
                            id_,
                            "+=" if name in used else "="))
                    else:
                        geo.append("Physical Line('{}', {}) {} {{{}}};".format(
                            name,
                            id_,
                            "+=" if name in used else "=",
                            nc.key))
                    used.add(name)

            for se in self.superelements:
                geo.append("Line Loop({}) = {{{}}};".format(
                    se.key,
                    ", ".join([str(nc.key) for nc in se.nodechains])))
                geo.append("Plane Surface({0}) = {{{0}}};".format(se.key))
            used = set()
            for se in self.superelements:
                id_ = physical_surface(se.elements[0]) - len(physical_lines)
                name = physical_surfaces[id_ - 1]
                if extrude:
                    geo.append(
                        "extrusion[] = Extrude {{0, 0, {}}} {{ Surface{{{}}}; {}{}}};".format(
                            extrude,
                            se.key,
                            "Layers{{{}}}; ".format(layers) if layers else "",
                            "Recombine; " if recombine else ""))
                    
                    geo.append("Physical Surface('base', {}) {} {{{}}};".format(
                        len(physical_lines) + 1,
                        "=" if se.key == 1 else "+=",
                        se.key))
                    
                    geo.append("Physical Surface('top', {}) {} extrusion[0];".format(
                        len(physical_lines) + 2,
                        "=" if se.key == 1 else "+="))
                    
                    geo.append("Physical Volume('{}', {}) {} extrusion[1];".format(
                        name,
                        id_,
                        "+=" if name in used else "="))
                else:
                    geo.append("Physical Surface('{}', {}) {} {{{}}};".format(
                        name,
                        id_,
                        "+=" if name in used else "=",
                        se.key))
                used.add(name)
            return geo

        if not self.FC_RADIUS:
            logger.warn("airgap radius is not set in source file")
        
        if fileformat == "msh":
            if extrude:
                raise ValueError("extrusion is only available for file format 'geo'")
            return get_msh()
        elif fileformat == "geo":
            return get_geo()
        raise ValueError("Invalid file format '{}'".format(fileformat))


class Point(object):
    def __init__(self, valid, x, y):
        self.valid = valid
        self.x = x
        self.y = y
        self.xy = x, y


class Line(object):
    def __init__(self, valid, p1, p2):
        self.valid = valid
        self.p1 = p1
        self.p2 = p2


class BaseEntity(object):
    def __init__(self, valid, key):
        self.valid = valid
        self.key = key


class Node(BaseEntity):
    def __init__(self, valid, key, bndcnd, pernod, x, y, vpot_re, vpot_im):
        super(self.__class__, self).__init__(valid, key)
        self.bndcnd = bndcnd
        self.pernod = pernod
        self.x = x
        self.y = y
        self.xy = x, y
        self.vpot = vpot_re, vpot_im

    def on_boundary(self):
        return self.bndcnd != 0 or self.pernod != 0


class NodeChain(BaseEntity):
    def __init__(self, valid, key, nodes):
        super(self.__class__, self).__init__(valid, key)
        self.node1 = nodes[0]
        self.nodemid = nodes[1]
        self.node2 = nodes[2]
        if nodes[1] is None:
            self.nodes = (nodes[0], nodes[2])
        else:
            self.nodes = (nodes[0], nodes[1], nodes[2])

    def reverse(self):
        return NodeChain(self.valid, self.key * (-1),
                         [self.node2, self.nodemid, self.node1])


class Element(BaseEntity):
    def __init__(self, valid, key, el_type, se_key, vertices, reluc, mag):
        super(self.__class__, self).__init__(valid, key)
        self.el_type = el_type
        self.se_key = se_key
        self.vertices = vertices
        self.reluc = reluc
        self.mag = mag


class SuperElement(BaseEntity):
    def __init__(self, valid, key, sr_key, elements, nodechains, color,
                 nc_keys, mcvtype, condtype, conduc, length,
                 velsys, velo_1, velo_2, curd_re, curd_im):
        super(self.__class__, self).__init__(valid, key)
        self.sr_key = sr_key
        self.elements = elements
        self.nodechains = nodechains
        self.color = color
        self.nc_keys = nc_keys
        self.mcvtype = mcvtype
        self.condtype = condtype
        self.conduc = conduc
        self.length = length
        self.velsys = velsys
        self.velo = velo_1, velo_2
        self.curd = curd_re, curd_im


class SubRegion(BaseEntity):
    def __init__(self, valid, key, sr_type, color, name, curdir, wb_key,
                 superelements, nodechains):
        super(self.__class__, self).__init__(valid, key)
        self.sr_type = sr_type
        self.color = color
        self.name = name
        self.curdir = curdir
        self.wb_key = wb_key
        self.superelements = superelements
        self.nodechains = nodechains


class Winding(BaseEntity):
    def __init__(self, valid, key, name, subregions, num_turns, cur_re, cur_im,
                 flux_re, flux_im, volt_re, volt_im):
        super(self.__class__, self).__init__(valid, key)
        self.name = name
        self.subregions = subregions
        self.num_turns = num_turns
        self.cur = cur_re, cur_im
        self.flux = flux_re, flux_im
        self.volt = volt_re, volt_im


def read(filename):
    """
    Read ISA7 file and return ISA7 object.

    Arguments:
        filename: name of I7/ISA7 file to be read
    """
    isa = Isa7(filename)
    return isa


def convert(infile, outfile, extrude=0):
    """
    Convert I7/ISA7 file to msh or geo format

    Arguments:
        infile: name of I7/ISA7 file to be converted
        outfile: name of converted file
        extrude: extrude surfaces using a translation along the z-axis
    """
    import os
    isa = Isa7(infile)
    fileformat = os.path.splitext(outfile)[-1].lower()
    
    with open(outfile, 'w') as f:
        if fileformat == '.geo':
            f.write('\n'.join(isa.to_gmsh("geo", extrude)))
        elif fileformat == '.msh':
            f.write('\n'.join(isa.to_gmsh("msh", extrude)))
        else:
            raise ValueError("Unkown File format for conversion: {}".format(
                fileformat))


if __name__ == "__main__":

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    else:
        filename = sys.stdin.readline().strip()

    isa = read(filename)



# -*- coding: utf-8 -*-
"""
    femagtools.opt
    ~~~~~~~~~~~~~~

    Manage multi-objective optimization with FEMAG



"""
import femagtools
import femagtools.fsl
import femagtools.moproblem
from .moo.algorithm import Nsga2
from .moo.population import Population

import logging

logger = logging.getLogger(__name__)


def log_pop(pop, ngen):
    objectives = pop.problem.objective_vars
    decisions = pop.problem.decision_vars
    log = [''.join(['Generation: {}\n'.format(ngen),
                    'rank '] +
                   ["{:10s}".format(o['name'].split('.')[-1])
                    for o in objectives] +
                   [" "] +
                   ["{:12s}".format(d['name'].split('.')[-1])
                    for d in decisions])]

    for i in pop.individuals:
        log.append(''.join(["{},".format(i.rank)] +
                           ["{:10.2f},".format(f) for f in i.cur_f] +
                           ["   "] +
                           ["{:10.4f},".format(x) for x in i.cur_x]))

    log.append(
        '--------------------------------------------------------------------')
    log.append("  znad: {}\n  zi: {}\n  zw: {}\n  Norm Dist: {}".format(
        pop.compute_nadir(),
        pop.compute_ideal(),
        pop.compute_worst(),
        pop.compute_norm_dist()))
    return log


class Optimizer(object):
    def __init__(self, workdir, magnetizingCurves, magnetMat):
        self.femag = femagtools.Femag(workdir,
                                      magnetizingCurves=magnetizingCurves,
                                      magnets=magnetMat)
        
    def _update_population(self, pop, engine):
        self.job.cleanup()

        for k, i in enumerate(pop.individuals):
            task = self.job.add_task()
            pop.problem.prepare(i.cur_x, self.model)
            for mc in self.femag.copy_magnetizing_curves(self.model,
                                                         task.directory):
                task.add_file(mc)
            task.add_file('femag.fsl',
                          self.builder.create(self.model, self.fea,
                                              self.femag.magnets))

        ntasks = engine.submit()
        status = engine.join()

        for t, i in zip(self.job.tasks, pop.individuals):
            if t.status == 'C':
                r = t.get_results()
                if isinstance(r, dict) and 'error' in r:
                    logger.warn("Task %s failed: %s", t.id, r['error'])
                else:
                    pop.problem.setResult(r)

                    i.cur_f = pop.problem.objfun([])
            else:
                logger.warn("Task %s failed with status %s", t.id, t.status)

        pop.update()

    def __call__(self, num_generations, opt, pmMachine,
                 operatingConditions, engine):
        return self.optimize(num_generations, opt, pmMachine,
                             operatingConditions, engine)
        
    def optimize(self, num_generations, opt, pmMachine,
                 operatingConditions, engine):
        """execute optimization"""
        decision_vars = opt['decision_vars']
        objective_vars = opt['objective_vars']
        population_size = opt['population_size']
                    
        problem = femagtools.moproblem.FemagMoProblem(decision_vars,
                                                      objective_vars)
        self.builder = femagtools.fsl.Builder()
        self.model = femagtools.model.MachineModel(pmMachine)
        self.fea = operatingConditions
        self.fea.update(self.model.windings)
        self.fea['lfe'] = self.model.lfe
        self.fea['move_action'] = self.model.move_action
        self.fea['pocfilename'] = (self.model.get('name') +
                                   '_' + str(self.model.get('poles')) +
                                   'p.poc')
        self.pop = Population(problem, population_size)
        
        algo = Nsga2()

        self.job = engine.create_job(self.femag.workdir)
        
        logger.info("Optimize x:%d f:%d generations:%d population size:%d",
                    len(self.pop.problem.decision_vars),
                    len(self.pop.problem.objective_vars),
                    num_generations,
                    self.pop.size())

        results = dict(rank=[], f=[], x=[])

        for i in range(num_generations):
            logger.info("Generation %d", i)
            if i > 0:
                newpop = algo.evolve(self.pop)
                self._update_population(newpop, engine)
                self.pop.merge(newpop)
            else:
                self._update_population(self.pop, engine)
            logger.info('\n'.join(log_pop(self.pop, i)))
        logger.info("finished")
        ft = []
        xt = []
        for i in self.pop.individuals:
            results['rank'].append(i.rank)
            ft.append(i.cur_f)
            xt.append(i.cur_x)
        objective_vars = self.pop.problem.objective_vars
        decision_vars = self.pop.problem.decision_vars
        results['f'] = [[s.get('sign', 1)*y for y in f]
                        for s, f in zip(objective_vars, zip(*ft))]
        results['x'] = list(zip(*xt))
        results['znad'] = [s.get('sign', 1)*v
                           for s, v in zip(objective_vars,
                                           self.pop.compute_nadir())]
        results['zi'] = [s.get('sign', 1)*v
                         for s, v in zip(objective_vars,
                                         self.pop.compute_ideal())]
        results['zw'] = [s.get('sign', 1)*v
                         for s, v in zip(objective_vars,
                                         self.pop.compute_worst())]
        results['dist'] = self.pop.compute_norm_dist()
        results['objective'] = [o['desc'] for o in objective_vars]
        results['decision'] = [d['desc'] for d in decision_vars]
        return results
    
    #print("\nChampion: {}\n".format(pop.champion['f']))
        #if flast != None:
        #    print("Fitness Comparison:")
        #    for f1, f2 in zip(pop.champion['f'], flast):
        #        print( "{:10.2f} {:10.2f}      {:10.2f}".format(f1, f2, f1-f2))
        #flast = list(pop.champion['f'])
        #print("")
#    except:
#        print "Unexpected error:", sys.exc_info()

#    print("L2 distance to the best decision vector:")
#    for best_decision in prob.best_x:
#        l2_norm = 0
#        for n in range(0, len(best_decision)):
#            l2_norm +=  (best_decision[n] - isl.population.champion.x[n]) ** 2
#        l2_norm = sqrt(l2_norm)
#        print(l2_norm)

#pf=pop.plot_pareto_fronts()
#savefig('pf0.png')



# -*- coding: utf-8 -*-
"""
    femagtools.erg
    ~~~~~~~~~~~~~~

    Reading ERG files



"""
import re
import numpy as np
import io


head_pattern = re.compile('([A-Za-z_0-9]+)')
unit_pattern = re.compile('\[([A-Za-z/0-9 ]+)\]')
num_pattern = re.compile(r'([+-]?\d+(?:\.\d+)?(?:[eE][+-]\d+)?)\s*')


def read(filename):
    """read ERG file
    returns dict with array values
    """
    head = []
    units = []
    m = []
    with io.open(filename,
                 errors='ignore') as f:
        for l in f:
            if head and units:
                n = num_pattern.findall(l)
                if n:
                    m.append([float(x) for x in n])
            elif head:
                u = unit_pattern.findall(l)
                if u:
                    units = u
            elif l.find('|') > 0:
                h = head_pattern.findall(l)
                if h:
                    head = h

    m = np.array(m).T
    ncols = len(set(m[1]))
    i1 = np.reshape(m[0], (-1, ncols)).T[0]
    nrows = len(i1)

    res = {k: (np.reshape(x, (nrows, ncols)).T[::-1]).tolist()
           for k, x in zip(head[2:], m[2:])}
    res['i1'] = i1.tolist()
    res['beta'] = m[1][:ncols][::-1].tolist()
    return res



# -*- coding: utf-8 -*-
"""
    dxfsl
    ~~~~~

    renders geometry


"""
import numpy as np
import matplotlib
import matplotlib.pyplot as pl
import matplotlib.patches as pch
from .shape import Shape
from .geom import convex_hull
import logging
import io

logger = logging.getLogger(__name__)
matplotlibversion = int(matplotlib.__version__.split('.')[0])
matplotlibrelease = int(matplotlib.__version__.split('.')[1])


#############################
#       PlotRenderer        #
#############################

class PlotRenderer(object):
    def __init__(self):
        self.fig = None
        self.background = '#eeeeee'
        pass

    def figure(self):
        if self.fig is None:
            self.fig = pl.figure(facecolor='lightblue',
                                 figsize=(9, 10))
            if matplotlibversion == 2 and matplotlibrelease > 0:
                pl.tight_layout(h_pad=0.2, w_pad=0.2)
            pl.subplots_adjust(bottom=0.05, top=0.95, hspace=0.25, wspace=0.15)
        return self.fig

    def add_plot(self, rows=1, cols=1, num=1):
        self.ax = self.figure().add_subplot(rows, cols,
                                            num,
                                            facecolor=self.background)
        return self.ax

    def add_emptyplot(self, rows=1, cols=1, num=1, title=''):
        ax = self.add_plot(rows, cols, num)
        ax.axis('off')
        if title:
            ax.set_title(title, size=14)

    def new_legend_handle(self, color, alpha, text):
        return pch.Patch(color=color, alpha=alpha, label=text)

    def show_plot(self):
        if self.fig is not None:
            pl.show()
            self.fig = None

    def arc(self, startangle, endangle, center, radius, color='blue'):
        self.ax.add_patch(pch.Arc(center, 2*radius, 2*radius,
                                  angle=0,
                                  theta1=startangle*180/np.pi,
                                  theta2=endangle*180/np.pi,
                                  color=color))

    def line(self, p1, p2, color='blue'):
        self.ax.add_line(pl.Line2D((p1[0], p2[0]),
                                   (p1[1], p2[1]), color=color))

    def circle(self, center, radius, color='blue'):
        self.ax.add_patch(pch.Circle(center, radius,
                                     fill=False, color=color))

    def point(self, p, col, color='blue'):
        pl.plot([p[0]], [p[1]], col, color=color)

    def fill(self, x, y, color, alpha):
        self.ax.fill(x, y, 'b', alpha=alpha, color=color, edgecolor='blue')

    def fill_circle(self, center, radius, color, alpha):
        circle = pl.Circle(center, radius, color=color, alpha=alpha)
        self.ax.add_artist(circle)
        circle = pl.Circle(center, radius, color='blue', fill=False)
        self.ax.add_artist(circle)

    def render(self, geom, filename=None, **kwargs):
        draw_center = kwargs.get('draw_center', False)
        draw_hull = kwargs.get('draw_hull', False)
        incl_bnd = kwargs.get('incl_bnd', False)

        fig = pl.figure()
        self.ax = fig.add_subplot(111)

        if draw_hull:
            for h in convex_hull(geom.g.nodes()):
                pl.plot([h[0]], [h[1]], 'ro')

        id = 0
        for area in geom.areas(incl_bnd):
            id += 1
            logger.info("area %d (size %d)", id, len(area))
            if len(area) > 1:
                for s in area:
                    s.render(self)
                    if draw_center:
                        c = s.center_of_connection()
                        pl.plot([c[0]], [c[1]], 'gs')

        geom.remove_areas(incl_bnd)

        # draw all remaining edges and circles
        [circle.render(self) for circle in geom.circles()]
        [attr['object'].render(self)
         for e1, e2, attr in geom.g.edges(data=True)]
        if draw_center:
            for e1, e2, attr in geom.g.edges(data=True):
                c = attr['object'].center_of_connection()
                pl.plot([c[0]], [c[1]], 'gs')

        geom.render_cut_lines(self)
        self.ax.axis('scaled', aspect='equal')
        if filename:
            pl.savefig(filename)
        else:
            pl.show()

    def print_convex_hull(self, nodes):
        points = list(nodes)
        points.sort()
        if len(points) <= 1:
            return points

        def cross(o, a, b):
            dx = a[0] - o[0], a[1] - o[1]
            dy = b[0] - o[0], b[1] - o[1]
            return dx[0] * dy[1] - dx[1] * dy[0]

        print("=== Print Convex Hull ===")
        lower = []
        for p in points:
            self.point(p, 'go')
            while len(lower) >= 2 and cross(lower[-2], lower[-1], p) <= 0:
                lower.pop()
            lower.append(p)
        print("lower={}".format(lower))

    def render_elements(self, geom, type, **kwargs):
        with_nodes = kwargs.get('with_nodes', False)
        with_hull = kwargs.get('with_hull', False)
        with_corners = kwargs.get('with_corners', False)
        with_center = kwargs.get('with_center', False)
        single_view = kwargs.get('single_view', False)
        neighbors = kwargs.get('neighbors', False)
        draw_center = kwargs.get('draw_center', False)
        draw_inside = kwargs.get('draw_inside', False)
        fill_areas = kwargs.get('fill_areas', False)
        title = kwargs.get('title', "")
        show = kwargs.get('show', True)
        rows = kwargs.get('rows', 1)
        cols = kwargs.get('cols', 1)
        num = kwargs.get('num', 1)
        points = kwargs.get('points', [])

        if show:
            rows = 1
            cols = 1
            num = 1

        mm = geom.minmax()
        x_min, x_max = mm[0]-5, mm[1]+5
        y_min, y_max = mm[2]-5, mm[3]+5

        if single_view:
            count = 0
            for e in geom.elements(type):
                print("Render Element {}".format(e))
                if count == 0:
                    self.ax = self.figure().add_subplot(111)
                    self.ax.axis('scaled', aspect='equal')
                    self.ax.set_xlim(x_min, x_max)
                    self.ax.set_ylim(y_min, y_max)
                e.render(self, 'blue', True)

                count += 1
                if count == 3:
                    pl.show()
                    count = 0

            if count != 0:
                self.show_plot()
            return

        self.ax = self.figure().add_subplot(rows, cols,
                                            num,
                                            facecolor=self.background)
        if len(title) > 0:
            self.ax.set_title(title, size=14)
        self.ax.grid(color='blue', linewidth=0.5)

        if with_hull:
            for h in convex_hull(geom.virtual_nodes()):
                pl.plot([h[0]], [h[1]], 'ro')

        if with_corners:
            for c in geom.start_corners:
                self.point(c, 'rs')
            for c in geom.end_corners:
                self.point(c, 'rs')

        if draw_center:
            for area in geom.list_of_areas():
                for s in area.elements():
                    c = s.center_of_connection()
                    pl.plot([c[0]], [c[1]], 'gs')

        if draw_inside:
            for area in geom.list_of_areas():
                p = area.get_point_inside(geom)
                if p:
                    pl.plot([p[0]], [p[1]], 'ro', color='magenta')

        if fill_areas:
            handles = geom.render_area_fill(self)
            if handles:
                legend = pl.legend(handles=handles, loc='best',
                                   fancybox=True,
                                   framealpha=1.0,
                                   fontsize=8,
                                   labelspacing=0.2)
                frame = legend.get_frame()
                frame.set_facecolor('white')
                frame.set_edgecolor('blue')

        for e in geom.elements(type):
            e.render(self, 'blue', with_nodes)

        geom.render_cut_lines(self)
        geom.render_airgaps(self)
        if neighbors:
            geom.render_neighbors(self)

        if len(points) > 0:
            for p in points:
                self.point(p, 'ro', color='red')

        if with_center and geom.center:
            self.point(geom.center, 'ro', color='darkgreen')
            x_min = min(x_min, geom.center[0]-5)
            x_max = max(x_max, geom.center[0]+5)
            y_min = min(y_min, geom.center[1]-5)
            y_max = max(y_max, geom.center[1]+5)

        self.ax.axis('scaled', aspect='equal')
        self.ax.set_xlim(x_min, x_max)
        self.ax.set_ylim(y_min, y_max)

        if show:
            self.show_plot()

    def render_areas(self, geom, **kwargs):
        with_nodes = kwargs.get('with_nodes', False)
        single_view = kwargs.get('single_view', False)
        title = kwargs.get('title', "")
        show = kwargs.get('show', True)
        rows = kwargs.get('rows', 1)
        cols = kwargs.get('cols', 1)
        num = kwargs.get('num', 1)

        if show:
            rows = 1
            cols = 1
            num = 1

        if not single_view:
            self.ax = self.figure().add_subplot(rows, cols, num)
            if len(title) > 0:
                self.ax.set_title(title, size=14)

        colors = ('red', 'green', 'blue', 'magenta',
                  'orange', 'grey', 'darkgreen')

        c = -1
        for area in geom.list_of_areas():
            if area.number_of_elements() > 1:
                c += 1
                if c >= len(colors):
                    c = 0
                if single_view:
                    print("*** AREA has {} elements ***".
                          format(area.number_of_elements()))
                    fig = pl.figure()
                    self.ax = fig.add_subplot(111)

                area.render(self, colors[c], with_nodes)
                p = area.get_point_inside(geom)
                if p:
                    self.point(p, 'ro', color='magenta')

                if single_view:
                    self.ax.axis('scaled', aspect='equal')
                    pl.show()

        if not single_view:
            self.ax.axis('scaled', aspect='equal')
            if show:
                self.show_plot()

    def render_area(self, area):
        fig = pl.figure()
        self.ax = fig.add_subplot(111)

        area.render(self, 'red', with_nodes=True)

        self.ax.axis('scaled', aspect='equal')
        pl.show()

    def draw_slot(self, id, slot, ax):
        poly = pch.Polygon(slot, fill=True, color='#ee82ee')

        center = (sum(list(zip(*slot))[0])/len(slot),
                  sum(list(zip(*slot))[1])/len(slot))
        ax.text(center[0], center[1], str(id))
        ax.add_patch(poly)


#############################
#       DumpRenderer        #
#############################

class DumpRenderer(object):
    def __init__(self, name):
        self.model = name

    def circle(self, center, radius):
        self.content.append(u'Crcle({}, {}, {})'.format(center[0],
                                                        center[1],
                                                        radius))

    def arc(self, startangle, endangle, center, radius):
        p1 = (center[0] + radius*np.cos(startangle),
              center[1] + radius*np.sin(startangle))
        p2 = (center[0] + radius*np.cos(endangle),
              center[1] + radius*np.sin(endangle))
        self.content.append(
            u"Arc({}, {}, {}, {}, {}, {})".format(
                p1[0], p1[1], p2[0], p2[1],
                center[0], center[1]))

    def line(self, p1, p2):
        self.content.append(
            u"Line({}, {}, {}, {})".format(
                p1[0], p1[1], p2[0], p2[1]))

    def render(self, geom):
        '''create file with elements'''
        incl_bnd = True
        self.content = []
        handled = set()

        # collect all areas
        for i, area in enumerate(geom.areas(incl_bnd)):
            self.content.append('-- {}'.format(i))
            for e in area:
                if e not in handled:
                    e.render(self)
                    handled.add(e)
        self.content.append('--')
        geom.remove_areas(incl_bnd)
        # and all remaining edges and circles
        for circle in geom.circles():
            circle.render(self)

        for e1, e2, attr in geom.g.edges(data=True):
            attr['object'].render(self)

        return self.content


#############################
#        FslRenderer        #
#############################

class FslRenderer(object):
    """a model that can created by FSL"""
    header = ['exit_on_error = false',
              'exit_on_end = false',
              'verbosity = 2',
              'pickdist = {}',
              'fm_nlin_colour=blue',
              'fm_nlin_color=fm_nlin_colour',
              'fm_nlin_mcvfile="M470P65A-AM-50Hz-96"',
              'fm_nlin_mcvfile_shft="ST52-3"',
              'fm_nlin_rlen = 100',
              'shaft_mat=1',
              'agndst = 0.5',
              'new_model_force("{}","Test")',
              'blow_up_wind(0,0, {}, {})']

    def __init__(self, name):
        self.model = name
        self.mirror_axis = None
        self.fm_nlin = None
        self.shaft = None

    def mirror_nodechains(self, p0, p1):
        self.mirror_axis = np.array((p0, p1)).ravel().tolist()

    def material(self, p0):
        self.fm_nlin = p0

    def circle(self, center, radius, color='blue'):
        num = int(2*np.pi*radius)
        if num < 8:
            num = 8
        circle = [u'cx, cy = {}, {}'.format(center[0],
                                            center[1]),
                  u'nc_circle_m(cx,cy,{}, {})'.format(radius, num),
                  u'create_mesh_se(cx, cy)\n']
        self.content += circle

    def arc(self, startangle, endangle, center, radius, color='blue'):
        num = 0
#        if self.nodedist > 0:
#            s = startangle
#            d = endangle - s
#            n = int(d/(2*np.pi))
#            if n < 0:
#                n -= 1
#            d -= n*2*np.pi
#            num = int(radius*d/self.nodedist + 1)
#            if num < 3 and radius*d > self.nodedist:
#                num = 3

        p1 = (center[0] + radius*np.cos(startangle),
              center[1] + radius*np.sin(startangle))
        p2 = (center[0] + radius*np.cos(endangle),
              center[1] + radius*np.sin(endangle))
        self.content.append(
            u"nc_circle_m({}, {}, {}, {}, {}, {}, {})".format(
                p1[0], p1[1], p2[0], p2[1],
                center[0], center[1], num))

    def line(self, p1, p2, color='blue'):
        num = 0
#        if self.nodedist > 0:
#            l = la.norm(np.asarray(p1)-p2)
#            num = int(l/self.nodedist + 1)
        self.content.append(
            u"nc_line({}, {}, {}, {}, {})".format(
                p1[0], p1[1], p2[0], p2[1], num))

    def sorted_elements(self, geom, inner=False):
        if inner:
            # Abstand von airgap Richtung Nullpunkt
            el_sorted = [(geom.max_radius -
                          e.minmax_from_center((0.0, 0.0))[1], e)
                         for e in geom.elements(Shape)]
        else:
            # Abstand von airgap Richtung Aussen
            el_sorted = [(e.minmax_from_center((0.0, 0.0))[0] -
                          geom.min_radius, e)
                         for e in geom.elements(Shape)]

        el_sorted.sort()
        return el_sorted

    def render(self, machine, filename, inner=False, outer=False):
        '''create file with nodechains'''

        geom = machine.geom
        self.content = []
        txt = [u'if agndst == nil or agndst == 0.0 then',
               u'  agndst = math.pi*(da1 + da2)/2/360',
               u'end',
               u'ndt(agndst)\n']
        self.content.append(u'\n'.join(txt))

        ndt_list = [(0.25, 1.25), (0.5, 2), (0.75, 3.0), (1.1, 3.0)]
        dist = geom.max_radius - geom.min_radius
        el_sorted = self.sorted_elements(geom, inner)

        x = 0
        for d, e in el_sorted:
            d_percent = d / dist
            if ndt_list[x][0] < d_percent:
                self.content.append(u'\nndt({}*agndst)\n'.
                                    format(ndt_list[x][1]))
                while x < 3 and ndt_list[x][0] < d_percent:
                    x += 1
#            self.content.append(u'-- d={} / dist={} == {}'.
#                                format(d, dist, d_percent))
            e.render(self)

        self.content.append(u'\n')

        parts = int(machine.get_symmetry_part())
        self.content.append(u'-- parts      = {}'
                            .format(parts))
        self.content.append(u'-- min_radius = {}'
                            .format(geom.min_radius))
        self.content.append(u'-- max_radius = {}'
                            .format(geom.max_radius))
        self.content.append(u'-- min_corner = {}, {}'
                            .format(geom.start_min_corner(0),
                                    geom.start_min_corner(1)))
        self.content.append(u'-- max_corner = {}, {}'
                            .format(geom.start_max_corner(0),
                                    geom.start_max_corner(1)))
        self.content.append(u'\n')

        if inner:
            self.content.append(u'tmp.inner_max_corner_x = {}'
                                .format(geom.start_max_corner(0)))
            self.content.append(u'tmp.inner_max_corner_y = {}'
                                .format(geom.start_max_corner(1)))
        if outer:
            self.content.append(u'tmp.outer_min_corner_x = {}'
                                .format(geom.start_min_corner(0)))
            self.content.append(u'tmp.outer_min_corner_y = {}'
                                .format(geom.start_min_corner(1)))
        self.content.append(u'\n')

        self.content.append(u'x0_iron_shaft, y0_iron_shaft = 0.0, 0.0')
        self.content.append(u'x0_iron_yoke, y0_iron_yoke = 0.0, 0.0\n')

        subregions = {}
        num_windings = 0
        num_magnets = 0
        for area in geom.list_of_areas():
            if area.number_of_elements() > 1:
                p = area.get_point_inside(geom)
                if p:
                    self.content.append(u"x0, y0 = {}, {}".format(p[0], p[1]))
                    self.content.append(u"point(x0, y0, red, 4)")
                    self.content.append(u"create_mesh_se(x0, y0)")

                if area.is_winding():
                    if area.type not in subregions:
                        subregions[area.type] = 1
                    num_windings += 1
                    self.content.append(u'm.xcoil_{}, m.ycoil_{} = x0, y0'.
                                        format(num_windings, num_windings))

                elif area.is_magnet():
                    if area.type not in subregions:
                        subregions[area.type] = 1
                    num_magnets += 1
                    self.content.append(u'tmp.xmag[{}], tmp.ymag[{}] = x0, y0'.
                                        format(num_magnets, num_magnets))
                    self.content.append(u'tmp.mag_orient[{}] = {}'.
                                        format(num_magnets, area.phi))
                elif area.type > 0:
                    if area.type in subregions:
                        self.content.append(
                            u'add_to_subreg(x0, y0, "{}")'.
                            format(area.name()))
                    else:
                        subregions[area.type] = 1
                        self.content.append(
                            u'def_new_subreg(x0, y0, "{}", {})'.
                            format(area.name(), area.color()))
                    if area.is_stator_iron_yoke():
                        self.content.append(
                            u'x0_iron_yoke, y0_iron_yoke = x0, y0')
                    if area.is_stator_iron_shaft():
                        self.content.append(
                            u'x0_iron_shaft, y0_iron_shaft = x0, y0')
                    if area.is_rotor_iron():
                        self.content.append(
                            u'x0_iron_yoke, y0_iron_yoke = x0, y0')

                self.content.append(u"\n")

        txt = [u"if x0_iron_yoke > 0.0 then",
               u"  if mcvkey_yoke ~= 'dummy' then",
               u'    def_mat_fm_nlin(x0_iron_yoke, y0_iron_yoke, blue, mcvkey_yoke, 100)',
               u'  else',
               u'    def_mat_fm(x0_iron_yoke, y0_iron_yoke, ur, 100)',
               u'  end',
               u'end\n']
        self.content.append(u'\n'.join(txt))

        txt = [u"if x0_iron_shaft > 0.0 then",
               u"  if mcvkey_shaft ~= 'dummy' then",
               u'    def_mat_fm_nlin(x0_iron_shaft, y0_iron_shaft, blue, mcvkey_shaft, 100)',
               u'  else',
               u'    def_mat_fm(x0_iron_shaft, y0_iron_shaft, ur, 100)',
               u'  end',
               u'end\n']
        self.content.append(u'\n'.join(txt))

        if num_windings > 0:
            self.content.append(u'tmp.coil_exists   = {}'.
                                format(num_windings))
            if geom.is_mirrored():
                self.content.append(u'tmp.coil_mirrored = true')
            else:
                self.content.append(u'tmp.coil_mirrored = false')
            self.content.append(u'tmp.coil_alpha    = {}'.format(geom.alfa))
            self.content.append(u'm.wdg_location  = 1.0 -- stator\n')

        if num_magnets > 0:
            self.content.append(u'tmp.mag_exists   = {}'.
                                format(num_magnets))
            if geom.is_mirrored():
                self.content.append(u'tmp.mag_mirrored = true')
            else:
                self.content.append(u'tmp.mag_mirrored = false')
            self.content.append(u'tmp.mag_alpha    = {}\n'
                                .format(geom.get_alfa()))

        if geom.is_mirrored():
            self.content.append(u'-- mirror')
            self.content.append(u'mirror_nodechains({}, {}, {}, {})\n'.format(
                geom.mirror_corners[1][0],   # max x1
                geom.mirror_corners[1][1],   # max y1
                geom.mirror_corners[0][0],   # min x2
                geom.mirror_corners[0][1]))  # min y2

        # Winkel nach allfälligem Spiegeln
        self.content.append(u'alfa = {}\n'.format(geom.get_alfa()))

        self.content.append(u'-- rotate')
        self.content.append(u'x1, y1 = {}, {}'.format(
                geom.start_corners[0][0], geom.start_corners[0][1]))  # min xy1
        self.content.append(u'x2, y2 = {}, {}'.format(
                geom.start_corners[1][0], geom.start_corners[1][1]))  # max xy2

        if geom.is_mirrored():
            self.content.append(u'x3, y3 = pr2c(x2, alfa)')
            self.content.append(u'x4, y4 = pr2c(x1, alfa)')
        else:
            self.content.append(u'x3, y3 = {}, {}'.format(
                    geom.end_corners[1][0], geom.end_corners[1][1]))  # max xy3
            self.content.append(u'x4, y4 = {}, {}'.format(
                    geom.end_corners[0][0], geom.end_corners[0][1]))  # min xy4

        self.content.append(u'if {} > 1 then'.format(geom.num_variable()))
        self.content.append(
            u'  rotate_copy_nodechains(x1,y1,x2,y2,x3,y3,x4,y4,{}-1)'
            .format(geom.num_variable()))
        self.content.append(u'end')
        self.content.append(u'alfa = {} * alfa'.format(geom.num_variable()))

        if self.fm_nlin:
            self.content.append(u'\nx0, y0 = {}, {}'. format(
                self.fm_nlin[0], self.fm_nlin[1]))

            mat = [u"if fm_nlin_mcvfile ~= 'dummy' then",
                   u"  if fm_nlin_mcvfile == 'air' then",
                   u"    def_mat_fm(x0,y0, 1.0, fm_nlin_rlen)",
                   u"  else",
                   u"    def_mat_fm_nlin(x0,y0, fm_nlin_colour, fm_nlin_mcvfile, fm_nlin_rlen)",
                   u"  end",
                   u"else",
                   u"  def_mat_fm(x0,y0, 1000.0, fm_nlin_rlen)",
                   u"end"]
            self.content.append(u'\n'.join(mat))

        if self.shaft:
            mat = [u'\nif shaft_mat==1 then',
                   u'  def_mat_fm_nlin(0.1,0.1,lightgrey,fm_nlin_mcvfile_shft,fm_nlin_rlen)',
                   u'end']
            self.content.append(u'\n'.join(mat))

        # f.write(u"\nadapt_window()\n")
        with io.open(filename, 'w', encoding='utf-8') as f:
            f.write('\n'.join(self.content))

    def render_main(self, motor, m_inner, m_outer,
                    filename, with_header=False):
        '''create main file'''

        self.content = []
        self.content.append(u'exit_on_error = false')
        self.content.append(u'exit_on_end = false')
        self.content.append(u'verbosity = 2\n')

        self.content.append(u'new_model_force("{}","Test")'.
                            format(self.model))
        self.content.append(u'global_unit(mm)')
        self.content.append(u'pickdist(0.001)')
        self.content.append(u'cosys(polar)\n')

        geom_inner = None
        geom_outer = None

        if m_inner and m_outer:
            geom_inner = m_inner.geom
            geom_outer = m_outer.geom

            parts_inner = int(m_inner.get_symmetry_part())
            parts_outer = int(m_outer.get_symmetry_part())

            if parts_inner > parts_outer:
                num_slots = parts_inner
                num_poles = parts_outer
                npols_gen = int(geom_outer.get_symmetry_copies()+1)
                num_sl_gen = int(geom_inner.get_symmetry_copies()+1)
            else:
                num_slots = parts_outer
                num_poles = parts_inner
                npols_gen = int(geom_inner.get_symmetry_copies()+1)
                num_sl_gen = int(geom_outer.get_symmetry_copies()+1)

            self.content.append(u'tmp = {}')
            self.content.append(u'tmp.xmag = {}')
            self.content.append(u'tmp.ymag = {}')
            self.content.append(u'tmp.mag_orient = {}')
            self.content.append(u'tmp.mag_exists = 0')
            self.content.append(u'tmp.coil_exists = 0\n')

            self.content.append(u'm.tot_num_slot = {}'.format(num_slots))
            self.content.append(u'm.num_sl_gen   = {}'.format(num_sl_gen))
            self.content.append(u'm.num_poles    = {}'.format(num_poles))
            self.content.append(u'm.npols_gen    = {}'.format(npols_gen))
            self.content.append(u'm.num_slots    = m.num_sl_gen')

        self.content.append(u'da1 = {}'.format(
            2*geom_outer.min_radius))
        self.content.append(u'da2 = {}'.format(
            2*geom_inner.max_radius))
        self.content.append(u'ag = (da1 - da2)/2\n')

        if m_inner and m_outer:
            self.content.append(u'm.tot_num_sl  = m.tot_num_slot')
            self.content.append(u'm.fc_radius   = (da1+da2)/4')
            self.content.append(u'm.fc_radius1  = m.fc_radius')
            self.content.append(u'pre_models("basic_modpar")\n')

        self.content.append(
            u'agndst = math.pi*m.fc_radius*m.npols_gen/m.num_poles/90')

        self.content.append(u'blow_up_wind(0, 0, 10, 10)\n')

        if geom_inner:
            self.content.append(u"mcvkey_yoke = 'dummy'")
            self.content.append(u"mcvkey_shaft = 'dummy'")
            self.content.append(u"ur = 1000.0")
            self.content.append(
                u'dofile("{}_{}.fsl")\n'.format(
                    self.model, geom_inner.kind))
        if geom_outer:
            self.content.append(u"mcvkey_yoke = 'dummy'")
            self.content.append(u"mcvkey_shaft = 'dummy'")
            self.content.append(u"ur = 1000.0")
            self.content.append(
                u'dofile("{}_{}.fsl")\n'.format(
                    self.model, geom_outer.kind))

        alfa = geom_inner.get_alfa() * (geom_inner.get_symmetry_copies()+1)
        alfa2 = geom_outer.get_alfa() * (geom_outer.get_symmetry_copies()+1)
        assert(np.isclose(alfa, alfa2))

        self.content.append(u'alfa = {}\n'.format(alfa))

        # Airgap
        txt = [u'-- airgap',
               u'ndt(agndst)',
               u'r1 = da2/2 + ag/3',
               u'x1, y1 = pr2c(r1, alfa)',
               u'n = r1*alfa/agndst + 1',
               u'nc_circle_m(r1, 0, x1, y1, 0.0, 0.0, n)\n',
               u'r2 = da2/2 + 2*ag/3',
               u'x2, y2 = pr2c(r2, alfa)',
               u'nc_circle_m(r2, 0, x2, y2, 0.0, 0.0, n)\n']
        self.content.append(u'\n'.join(txt))
        self.content.append(u'x1, y1 = {}, {}'
                            .format(geom_inner.start_max_corner(0),
                                    geom_inner.start_max_corner(1)))
        self.content.append(u'nc_line(x1, y1, r1, 0.0, 0.0)\n')
        self.content.append(u'x2, y2 = {}, {}'
                            .format(geom_outer.start_min_corner(0),
                                    geom_outer.start_min_corner(1)))
        self.content.append(u'nc_line(r2, 0.0, x2, y2, 0.0)\n')

        txt = [u'x3, y3 = pr2c(x1, alfa)',
               u'x4, y4 = pr2c(r1, alfa)',
               u'nc_line(x3, y3, x4, y4, 0, 0)\n',
               u'x3, y3 = pr2c(x2, alfa)',
               u'x4, y4 = pr2c(r2, alfa)',
               u'nc_line(x3, y3, x4, y4, 0, 0)\n',
               u'x0, y0 = pr2c(r1-ag/6, alfa/2)',
               u'create_mesh_se(x0, y0)',
               u'x0, y0 = pr2c(r2+ag/6, alfa/2)',
               u'create_mesh_se(x0, y0)\n']
        self.content.append(u'\n'.join(txt))

        self.content.append(u'connect_models()\n')

        # Windings
        txt = [u'-- Gen_winding',
               u'if tmp.coil_exists > 0 then',
               u'  m.num_phases      = 3',
               u'  m.num_layers      = 2',
               u'  m.num_wires       = 1',
               u'  m.coil_span       = 1',
               u'  m.current         = 0.0',
               u'  m.mat_type        = 1.0 -- rotating',
               u'  m.wind_type       = 1.0 -- winding & current',
               u'  m.win_asym        = 1.0 -- sym',
               u'  m.wdg_location    = 1.0 -- stator',
               u'  m.curr_inp        = 0.0 -- const',
               u'  m.dq_offset       = 0',
               u'  if tmp.coil_exists == 1 and tmp.coil_mirrored then',
               u'    r, phi = c2pr(m.xcoil_1, m.ycoil_1)',
               u'    m.xcoil_2, m.ycoil_2 = pr2c(r, tmp.coil_alpha*2.0 - phi)',
               u'  end\n',
               u'  pre_models("Gen_winding")',
               u'end\n']
        self.content.append(u'\n'.join(txt))

        txt = [u'-- pm magnets',
               u'if tmp.mag_exists > 0 then',
               u'  alfa = tmp.mag_alpha',
               u'  m.remanenc = 1.15',
               u'  m.relperm  = 1.05',
               u'  for i=0, m.npols_gen-1 do',
               u'    for n=1, tmp.mag_exists do',
               u'      r, p = c2pr(tmp.xmag[n], tmp.ymag[n])',
               u'      phi = i*alfa+p',
               u'      x0, y0 = pr2c(r, phi)',
               u'      phi_orient = i*alfa+tmp.mag_orient[n]',
               u'      orient = phi_orient*180/math.pi',
               u'      if ( i % 2 == 0 ) then',
               u'        def_mat_pm(x0, y0, red, m.remanenc, m.relperm,',
               u'                   orient-180, m.parallel, 100)',
               u'      else',
               u'        def_mat_pm(x0, y0, green, m.remanenc, m.relperm,',
               u'                   orient, m.parallel, 100)',
               u'      end',
               u'      if tmp.mag_mirrored then',
               u'        phi = (i+1)*alfa-p',
               u'        x0, y0 = pr2c(r, phi)',
               u'        phi_orient = (i+1)*alfa-tmp.mag_orient[n]',
               u'        orient = phi_orient*180/math.pi',
               u'        if ( i % 2 == 0 ) then',
               u'          def_mat_pm(x0, y0, red, m.remanenc, m.relperm,',
               u'                     orient-180, m.parallel, 100)',
               u'        else',
               u'          def_mat_pm(x0, y0, green, m.remanenc, m.relperm,',
               u'                     orient, m.parallel, 100)',
               u'        end',
               u'      end',
               u'    end',
               u'  end',
               u'end']
        self.content.append(u'\n'.join(txt))

        with io.open(filename, 'w', encoding='utf-8') as f:
            f.write('\n'.join(self.content))



# -*- coding: utf-8 -*-
"""manage a geometry with
    lines, circles and arcs built from DXF

  NOTE: This code is in highly experimental state.
        Use at your own risk.

  Author: Ronald Tanner
    Date: 2017/07/06
"""
from __future__ import print_function
import numpy as np
import logging
from .functions import distance

logger = logging.getLogger('femagtools.corner')


#############################
#           Corner          #
#############################

class Corner(object):
    def __init__(self, center, p):
        self.__p = p
        self.__dist = distance(center, p)
        self.__keep = False
        self.is_new_point = False

    def point(self):
        return self.__p

    def is_equal(self, p, rtol=1e-04, atol=1e-04):
        return (np.isclose(self.__p[0], p[0], rtol=rtol, atol=atol) and
                np.isclose(self.__p[1], p[1], rtol=rtol, atol=atol))

    def is_same_corner(self, c):
        return self.is_equal(c.__p)

    def set_keep_node(self):
        self.__keep = True

    def keep_node(self):
        return self.__keep

    def __lt__(self, c):
        if self.__dist < c.__dist:
            return 1
        else:
            return 0

    def __str__(self):
        return "Corner: p={}".format(self.__p)



# -*- coding: utf-8 -*-
"""
    femagtools
    ~~~~~~~~~~

    Python bindings for FEMAG



"""
__title__ = 'femagtools'
__version__ = '0.5.1'
__author__ = 'Ronald Tanner'
__license__ = 'BSD'
__copyright__ = 'Copyright 2016 SEMAFOR Informatik & Energie AG'

from .bch import Reader
from .model import MachineModel
from .fsl import Builder
from .magnet import Magnet
from .femag import Femag, ZmqFemag


def read_bchfile(filename):
    """Read BCH/BATCH results from file *filename*."""
    import io
    bchresults = Reader()
    with io.open(filename, encoding='latin1', errors='ignore') as f:
        bchresults.read(f.readlines())
    return bchresults


def create_fsl(machine,
               operatingconditions={},
               magnetmat=[]):
    """create FSL command list from model parameters
    
    Args:
        machine: dict with parameters
        operatuingConditions: dict with parameters
        magnetmat: list fo dict with parameters
"""
    model = MachineModel(machine)
    
    builder = Builder()
    magnets = []
    if magnetmat:
        magnets = Magnet(magnetmat)
    if operatingconditions:
            return builder.create(model, operatingconditions, magnets)
    return builder.create_model(model, magnets) + ['save_model(cont)']
    
